import sys
import logging
from . import gcp_cis_utils as gcp_cis_utils

def check2_1():
    # 2.1 Ensure that Cloud Audit Logging is configured properly across all services and all users from a project (Scored)

    logging.info("2.1 Ensure that Cloud Audit Logging is configured properly across all services and all users from a project (Scored)")
    details = []
    iam_policies_by_project = gcp_cis_utils.get_iam_policies_by_projects()
    for p in iam_policies_by_project.keys():
        out_json = iam_policies_by_project[p]
        acs = out_json.get('auditConfigs')
        if acs is None:
            details.append("Cloud Audit Logging is not configured for project [%s]" % p)
            continue
        for ac in acs:
            s = ac.get('service')
            if s is None or s != "allServices":
                details.append("Cloud Audit Logging is not configured for all services for project [%s]" % p)
                continue
            alcs = ac.get('auditLogConfigs')
            if alcs is None:
                details.append("Cloud Audit Logging is not configured properly for project [%s]" % p)
                continue
            lt_dr_found = False
            lt_dw_found = False
            for alc in alcs:
                if alc.get('logType') == "DATA_READ":
                    lt_dr_found = True
                elif alc.get('logType') == "DATA_WRITE":
                    lt_dw_found = True
                ems = alc.get('exemptedMembers')
                if ems is None:
                    continue
                for em in ems:
                    if em.startswith('user:'):
                        details.append("Audit configuration has exempted member [%s] for log type [%s] for project [%s]" % (em.split(':')[1], alc.get('logType'), p))
            if lt_dr_found == False:
                details.append("Cloud audit logging configuration is not enabled for DATA_READ operation for  project [%s]" % p)
            if lt_dw_found == False:
                details.append("Cloud audit logging configuration is not enabled for DATA_WRITE operation for  project [%s]" % p)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.1', '2.1 [Level 1] Ensure that Cloud Audit Logging is configured properly across all services and all users from a project (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_2():
    # 2.2 Ensure that sinks are configured for all log entries (Scored)

    logging.info("2.2 Ensure that sinks are configured for all log entries (Scored)")
    details = []
    # Check if sink is configured at organization or folder level
    orgs = gcp_cis_utils.get_all_organizations()
    for o in orgs:
        out_json = gcp_cis_utils.run_gcloud_cmd("logging sinks list --organization=%s" % o)
        for entry in out_json:
            if entry['filter'] == "(empty filter)":
                # TODO confirm that the destination exits
                return None
    folders = gcp_cis_utils.get_all_folders()
    for f in folders:
        out_json = gcp_cis_utils.run_gcloud_cmd("logging sinks list --folder=%s" % f)
        for entry in out_json:
            if entry['filter'] == "(empty filter)":
                # TODO confirm that the destination exits
                return None
    projects = gcp_cis_utils.get_logging_enabled_projects()
    for p in projects:
        sink_found = False
        out_json = gcp_cis_utils.run_gcloud_cmd("logging sinks list --project=%s" % p)
        for entry in out_json:
            if entry['filter'] == "(empty filter)":
                # TODO confirm that the destination exits
                sink_found = True
        if sink_found == False:
            details.append("Sinks are not configured for all log entries for project [%s]" % p)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.2', '2.2 [Level 1] Ensure that sinks are configured for all log entries (Scored)', "\n".join(details), '4', '', '')
    return None

def _check_bucket_retention_policy(logbucket, details):
    output = gcp_cis_utils.run_cmd("gsutil retention get %s 2>/dev/null" % logbucket)
    if "has no Retention Policy" in output:
        details.append("No retention policy is configured for log bucket [%s]" % logbucket)
    if "Retention Policy (UNLOCKED)" in output:
        details.append("Retention policy is not locked for log bucket [%s]" % logbucket)

def check2_3():
    # 2.3 Ensure that retention policies on log buckets are configured using Bucket Lock (Scored)

    logging.info("2.3 Ensure that retention policies on log buckets are configured using Bucket Lock (Scored)")
    details = []
    # Get sinks for organizations
    orgs = gcp_cis_utils.get_all_organizations()
    for o in orgs:
        out_json = gcp_cis_utils.run_gcloud_cmd("logging sinks list --organization=%s" % o)
        for entry in out_json:
            if entry['destination'].startswith("storage.googleapis.com/"):
                logbucket = "gs://" + entry['destination'][23:]
                _check_bucket_retention_policy(logbucket, details)
    # Get sinks for folders
    folders = gcp_cis_utils.get_all_folders()
    for f in folders:
        out_json = gcp_cis_utils.run_gcloud_cmd("logging sinks list --folder=%s" % f)
        for entry in out_json:
            if entry['destination'].startswith("storage.googleapis.com/"):
                logbucket = "gs://" + entry['destination'][23:]
                _check_bucket_retention_policy(logbucket, details)
    # Get sinks for projects
    projects = gcp_cis_utils.get_logging_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("logging sinks list --project=%s" % p)
        for entry in out_json:
            if entry['destination'].startswith("storage.googleapis.com/"):
                logbucket = "gs://" + entry['destination'][23:]
                _check_bucket_retention_policy(logbucket, details)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.3', '2.3 [Level 1] Ensure that retention policies on log buckets are configured using Bucket Lock (Scored)', "\n".join(details), '4', '', '')
    return None

def _check_log_metric_filter_and_alerts(in_filter, details_msg):
    details = []
    projects = gcp_cis_utils.get_logging_enabled_projects()
    for p in projects:
        out_json = gcp_cis_utils.run_gcloud_cmd("beta logging metrics list --project=%s" % p)
        poac_metric_found = False
        for entry in out_json:
            mf = entry['filter'].strip().replace('\n',' ')
            if mf == in_filter:
                mdt = entry['metricDescriptor']['type']
                out_json_2 = gcp_cis_utils.run_gcloud_cmd("alpha monitoring policies list --project=%s" % p)
                for entry_2 in out_json_2:
                    for cond in entry_2['conditions']:
                        if cond['conditionThreshold']['filter'] == "metric.type=\"" + mdt + "\"" and entry_2['enabled']:
                            poac_metric_found = True
        if poac_metric_found == False:
            details.append(details_msg % p)
    return details

def check2_4():
    # 2.4 Ensure log metric filter and alerts exist for project ownership assignments/changes (Scored)

    logging.info("2.4 Ensure log metric filter and alerts exist for project ownership assignments/changes (Scored)")
    POAC_METRIC_FILTER = '(protoPayload.serviceName="cloudresourcemanager.googleapis.com") AND (ProjectOwnership OR projectOwnerInvitee) OR (protoPayload.serviceData.policyDelta.bindingDeltas.action="REMOVE" AND protoPayload.serviceData.policyDelta.bindingDeltas.role="roles/owner") OR (protoPayload.serviceData.policyDelta.bindingDeltas.action="ADD" AND protoPayload.serviceData.policyDelta.bindingDeltas.role="roles/owner")'
    details_msg = "Log metric filter and alerts do not exist for Project Ownership assignments/changes for project [%s]"
    details =  _check_log_metric_filter_and_alerts(POAC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.4', '2.4 [Level 1] Ensure log metric filter and alerts exist for project ownership assignments/changes (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_5():
    # 2.5 Ensure that the log metric filter and alerts exist for Audit Configuration changes (Scored)

    logging.info("2.5 Ensure that the log metric filter and alerts exist for Audit Configuration changes (Scored)")
    ACC_METRIC_FILTER = 'protoPayload.methodName="SetIamPolicy" AND protoPayload.serviceData.policyDelta.auditConfigDeltas:*'
    details_msg = "Log metric filter and alerts do not exist for Audit Configuration changes for project [%s]"
    details = _check_log_metric_filter_and_alerts(ACC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.5', '2.5 [Level 1] Ensure that the log metric filter and alerts exist for Audit Configuration changes (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_6():
    # 2.6 Ensure that the log metric filter and alerts exist for Custom Role changes (Scored)

    logging.info("2.6 Ensure that the log metric filter and alerts exist for Custom Role changes (Scored)")

    CRC_METRIC_FILTER = 'resource.type="iam_role" AND protoPayload.methodName = "google.iam.admin.v1.CreateRole" OR protoPayload.methodName="google.iam.admin.v1.DeleteRole" OR protoPayload.methodName="google.iam.admin.v1.UpdateRole"'
    details_msg = "Log metric filter and alerts do not exist for Custom Role changes for project [%s]"
    details = _check_log_metric_filter_and_alerts(CRC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.6', '2.6 [Level 1] Ensure that the log metric filter and alerts exist for Custom Role changes (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_7():
    # 2.7 Ensure that the log metric filter and alerts exist for VPC Network Firewall rule changes (Scored)

    logging.info("2.7 Ensure that the log metric filter and alerts exist for VPC Network Firewall rule changes (Scored)")

    VPCNFRC_METRIC_FILTER = 'resource.type="gce_firewall_rule" AND jsonPayload.event_subtype="compute.firewalls.patch" OR jsonPayload.event_subtype="compute.firewalls.insert"'
    details_msg = "Log metric filter and alerts do not exist for VPC Network Firewall Rule changes for project [%s]"
    details = _check_log_metric_filter_and_alerts(VPCNFRC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.7', '2.7 [Level 1] Ensure that the log metric filter and alerts exist for VPC Network Firewall rule changes (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_8():
    # 2.8 Ensure that the log metric filter and alerts exist for VPC network route changes (Scored)

    logging.info("2.8 Ensure that the log metric filter and alerts exist for VPC network route changes (Scored)")

    VPCNRC_METRIC_FILTER = 'resource.type="gce_route" AND jsonPayload.event_subtype="compute.routes.delete" OR jsonPayload.event_subtype="compute.routes.insert"'
    details_msg = "Log metric filter and alerts do not exist for VPC Network Route changes for project [%s]"
    details = _check_log_metric_filter_and_alerts(VPCNRC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.8', '2.8 [Level 1] Ensure that the log metric filter and alerts exist for VPC network route changes (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_9():
    # 2.9 Ensure that the log metric filter and alerts exist for VPC network changes (Scored)

    logging.info("2.9 Ensure that the log metric filter and alerts exist for VPC network changes (Scored)")

    VPCNC_METRIC_FILTER = 'resource.type=gce_network AND jsonPayload.event_subtype="compute.networks.insert" OR jsonPayload.event_subtype="compute.networks.patch" OR jsonPayload.event_subtype="compute.networks.delete" OR jsonPayload.event_subtype="compute.networks.removePeering" OR jsonPayload.event_subtype="compute.networks.addPeering"'
    details_msg = "Log metric filter and alerts do not exist for VPC Network changes for project [%s]"
    details = _check_log_metric_filter_and_alerts(VPCNC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.9', '2.9 [Level 1] Ensure that the log metric filter and alerts exist for VPC network changes (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_10():
    # 2.10 Ensure that the log metric filter and alerts exist for Cloud Storage IAM permission changes (Scored)

    logging.info("2.10 Ensure that the log metric filter and alerts exist for Cloud Storage IAM permission changes (Scored)")

    CSIAMPC_METRIC_FILTER = 'resource.type=gcs_bucket AND protoPayload.methodName="storage.setIamPermissions"'
    details_msg = "Log metric filter and alerts do not exist for Cloud Storage IAM Permission changes for project [%s]"
    details = _check_log_metric_filter_and_alerts(CSIAMPC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.10', '2.10 [Level 1] Ensure that the log metric filter and alerts exist for Cloud Storage IAM permission changes (Scored)', "\n".join(details), '4', '', '')
    return None

def check2_11():
    # 2.11 Ensure that the log metric filter and alerts exist for SQL instance configuration changes (Scored)

    logging.info("2.11 Ensure that the log metric filter and alerts exist for SQL instance configuration changes (Scored)")

    SQLICC_METRIC_FILTER = 'protoPayload.methodName="cloudsql.instances.update"'
    details_msg = "Log metric filter and alerts do not exist for SQL Instance Configuration changes for project [%s]"
    details = _check_log_metric_filter_and_alerts(SQLICC_METRIC_FILTER, details_msg)
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-2.11', '2.11 [Level 1] Ensure that the log metric filter and alerts exist for SQL instance configuration changes (Scored)', "\n".join(details), '4', '', '')
    return None

def run_checks():
    config_issues = []
    gcp_cis_utils.append_issue(config_issues, check2_1())
    gcp_cis_utils.append_issue(config_issues, check2_2())
    gcp_cis_utils.append_issue(config_issues, check2_3())
    gcp_cis_utils.append_issue(config_issues, check2_4())
    gcp_cis_utils.append_issue(config_issues, check2_5())
    gcp_cis_utils.append_issue(config_issues, check2_6())
    gcp_cis_utils.append_issue(config_issues, check2_7())
    gcp_cis_utils.append_issue(config_issues, check2_8())
    gcp_cis_utils.append_issue(config_issues, check2_9())
    gcp_cis_utils.append_issue(config_issues, check2_10())
    gcp_cis_utils.append_issue(config_issues, check2_11())
    return config_issues




import os
import socket
import subprocess
import paramiko
import logging

def run_cmd_on_host(args, host, cmdarr, logging_enabled=True):
    if host and host['remote']:
        pkgout = run_remote_ssh_command(args, host, cmdarr[0])
        if pkgout is None:
            return None
    else:
        try:
            dev_null_device = open(os.devnull, "w")
            pkgout = subprocess.check_output(cmdarr, stderr=dev_null_device, shell=True)
            pkgout = pkgout.decode(args.encoding)
            dev_null_device.close()
        except subprocess.CalledProcessError:
            if logging_enabled:
                logging.error("Error running inventory")
            return None
    return pkgout

def run_remote_ssh_command(args, host, command):
    assetid = host['assetid'] if host.get('assetid') is not None else host['hostname']
    output = ''
    try:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy)
        if host.get('userpwd') is not None and len(host['userpwd']) > 0 and (host.get('privatekey') is None or len(host['privatekey'])==0):
            client.connect(host['hostname'],username=host['userlogin'],password=host['userpwd'])
        elif host.get('privatekey') is not None and len(host['privatekey']) > 0:
            if host.get('userpwd') is not None and len(host['userpwd']) > 0:
                client.connect(host['hostname'],username=host['userlogin'],key_filename=host['privatekey'],passphrase=host['userpwd'])
            else:
                client.connect(host['hostname'],username=host['userlogin'],key_filename=host['privatekey'])
        else:
            client.connect(host['hostname'],username=host['userlogin'])
        stdin, stdout, stderr = client.exec_command(command)
        for line in stdout:
            output = output + line
        client.close()
    except paramiko.ssh_exception.AuthenticationException as e:
        logging.info("Authentication failed for asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except paramiko.ssh_exception.SSHException as e:
        logging.info("SSHException while connecting to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except socket.error as e:
        logging.info("Socket error while connection to asset [%s], host [%s]", assetid, host['hostname'])
        logging.info("Exception: %s", e)
        output = None
    except:
        logging.info("Unknown error running remote discovery for asset [%s], host [%s]: [%s]", assetid, host['hostname'], sys.exc_info()[0])
        output = None
    finally:
        return output

def get_os_release(args, host=None):
    freebsd = False
    out = None
    cmdarr = ["/bin/cat /etc/os-release"]
    out = run_cmd_on_host(args, host, cmdarr, False)

    if out is None or out.strip() == '':

        # try redhat-release
        cmdarr = ["/bin/cat /etc/redhat-release"]
        out = run_cmd_on_host(args, host, cmdarr, False)
        if out is not None and out.strip() != '':
            return out.strip()
        else:
            # try FreeBSD
            cmdarr = ["/usr/bin/uname -v -p"]
            out = run_cmd_on_host(args, host, cmdarr, False)

            if out is not None and 'FreeBSD' not in out:
                # try OpenBSD
                cmdarr = ["/usr/bin/uname -srvm"]
                out = run_cmd_on_host(args, host, cmdarr, False)

    if out is None:
        logging.error("Failed to get os-release")
        return None

    if 'FreeBSD' in out or 'OpenBSD' in out:
        return out
    elif 'Darwin' in out:
        # Check for Mac OS
        cmdarr = ["sw_vers"]
        out = run_cmd_on_host(args, host, cmdarr, False)
        if out is not None and out.strip() != '':
            mac_os_version = ""
            for line in out.splitlines():
                value = line.split(':')[1].strip()
                mac_os_version = value if len(mac_os_version) == 0 else mac_os_version + " " + value
            return mac_os_version
    else:
        output_lines = out.splitlines()
        for l in output_lines:
            if 'PRETTY_NAME' in l:
                return l.split('=')[1].replace('"','')
    return None


def get_asset_type(os):
    if "CentOS" in os:
        return "CentOS"
    elif "Red Hat" in os:
        return "Red Hat"
    elif "Ubuntu" in os:
        return "Ubuntu"
    elif "Debian" in os:
        return "Debian"
    elif "Amazon Linux" in os:
        return "Amazon Linux"
    elif "Oracle Linux" in os:
        return "Oracle Linux"
    elif "FreeBSD" in os:
        return "FreeBSD"
    elif "OpenBSD" in os:
        return "OpenBSD"
    elif "Mac OS" in os:
        return "Mac OS"
    elif "Windows" in os:
        return "Windows"
    else:
        logging.error("Not a supported OS type [%s]" % os)
        return None


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

def find_files(localpath, filename):
    ret_files = []
    for root, subdirs, files in os.walk(localpath):
        for fname in files:
            file_path = os.path.join(root, fname)
            if len(filename) == 0:
                ret_files.append(file_path)
            elif file_path.endswith(filename):
                ret_files.append(file_path)
    return ret_files

def ascii_string(in_str):
    ascii_str = ''.join([c if ord(c) < 128 else ' ' for c in in_str.strip()])
    return ascii_str

def get_indent(line):
    return len(line) - len(line.lstrip(' '))

def get_rating(cvss_score):
    rating = '1'
    if cvss_score is None or cvss_score == '':
        return rating
    s = float(cvss_score)
    if s <= 2:
        rating = '1'
    elif s <= 4:
        rating = '2'
    elif s <= 6:
        rating = '3'
    elif s <= 8:
        rating = '4'
    elif s <= 10:
        rating = '5'
    return rating



import sys
import subprocess
import json
import logging

# Some global variables here
_projects = None
_organizations = None
_folders = None
_iam_policies_by_projects = None
_compute_instances_by_projects = None
_services_by_projects = None

_encoding = None

def set_encoding(encoding):
    global _encoding
    _encoding = encoding

def get_encoding():
    global _encoding
    return _encoding

def run_cmd(cmd):
    try:
        cmd_output = subprocess.check_output([cmd], shell=True, stdin=None, stderr=None)
        cmd_output = cmd_output.decode(get_encoding())
    except subprocess.CalledProcessError:
        logging.error("Error running command [%s]", cmd)
        cmd_output = ""
    return cmd_output

def run_gcloud_cmd(cmd):
    cmd = 'gcloud ' + cmd + ' --format=json --quiet'
    try:
        cmd_output = subprocess.check_output([cmd], shell=True, stdin=None, stderr=None)
        cmd_output = cmd_output.decode(get_encoding())
        ret_json = json.loads(cmd_output)
    except subprocess.CalledProcessError:
        logging.error("Error running gcloud command [%s]", cmd)
        ret_json = { }
    except ValueError:
        logging.error("Error parsing JSON output for gcloud command [%s]: %s", cmd, cmd_output)
        ret_json = { }
    return ret_json

def create_issue(twc_id, twc_title, details, rating, object_id, object_meta):
    issue = { }
    issue['twc_id'] = twc_id
    issue['twc_title'] = twc_title
    issue['details'] = details
    issue['rating'] = rating
    issue['object_id'] = object_id
    issue['object_meta'] = object_meta
    return issue

def append_issue(config_issues, issue):
    if issue is not None:
        config_issues.append(issue)

# Get all projects
def get_all_projects():
    global _projects
    if _projects is not None:
        return _projects
    _projects = []
    out_json = run_gcloud_cmd('projects list')
    for p in out_json:
        _projects.append(p['projectId'])
    return _projects

# Get all organizations
def get_all_organizations():
    global _organizations
    if _organizations is not None:
        return _organizations
    _organizations = []
    out_json = run_gcloud_cmd('organizations list')
    for o in out_json:
        _organizations.append(o['organizationId'])
    return _organizations

def _add_sub_folders(folder, folders):
    out_json = run_gcloud_cmd("resource-manager folders list --folder=%s" % folder)
    for f in out_json:
        folders.append(f['folderId'])
        add_sub_folders(f['folderId'], folders)

# List all folders
def get_all_folders():
    global _folders
    if _folders is not None:
        return _folders
    _folders = []
    orgs = get_all_organizations()
    for o in orgs:
        out_json = run_gcloud_cmd("resource-manager folders list --organization=%s" % o)
        for f in out_json:
            folders.append(f['folderId'])
            # Get sub-folders if any
            _add_sub_folders(f['folderId'], folders)
    return _folders

def get_iam_policies_by_projects():
    global _iam_policies_by_projects
    if _iam_policies_by_projects is not None:
        return _iam_policies_by_projects
    _iam_policies_by_projects = { }
    projects = get_all_projects()
    for p in projects:
        out_json = run_gcloud_cmd("projects get-iam-policy %s" % p)
        _iam_policies_by_projects[p] = out_json
    return _iam_policies_by_projects

def get_compute_instances_by_projects():
    global _compute_instances_by_projects
    if _compute_instances_by_projects is not None:
        return _compute_instances_by_projects
    _compute_instances_by_projects = { }
    projects = get_compute_enabled_projects()
    for p in projects:
        out_json = run_gcloud_cmd("compute instances list --project=%s" % p)
        _compute_instances_by_projects[p] = out_json
    return _compute_instances_by_projects

def add_members_with_role(entry, role, add_to_set):
    if entry['role'] == role:
        for m in entry['members']:
            if m.startswith('user:'):
                add_to_set.add(m.split(':')[1])

def _get_services_by_projects():
    global _services_by_projects
    if _services_by_projects is not None:
        return _services_by_projects
    _services_by_projects = { }
    projects = get_all_projects()
    for p in projects:
        out_json = run_gcloud_cmd("services list --project=%s" % p)
        _services_by_projects[p] = out_json
    return _services_by_projects

def _get_projects_with_service_enabled(service):
    ret_projects = []
    services_by_projects = _get_services_by_projects()
    for p in services_by_projects.keys():
        out_json = services_by_projects[p]
        for entry in out_json:
            if entry['name'].endswith(service) and entry['state'] == "ENABLED":
                ret_projects.append(p)
    return ret_projects

def get_cloud_kms_enabled_projects():
    return _get_projects_with_service_enabled("cloudkms.googleapis.com")

def get_logging_enabled_projects():
    return _get_projects_with_service_enabled("logging.googleapis.com")

def get_compute_enabled_projects():
    return _get_projects_with_service_enabled("compute.googleapis.com")

def get_dns_enabled_projects():
    return _get_projects_with_service_enabled("dns.googleapis.com")

def get_bigquery_enabled_projects():
    return _get_projects_with_service_enabled("bigquery.googleapis.com")

def check_database_flag(dbtype, flag, value, details_msg):
    details = []
    projects = get_all_projects()
    for p in projects:
        out_json = run_gcloud_cmd("sql instances list --project=%s --filter='DATABASE_VERSION:%s'" % (p, dbtype))
        for entry in out_json:
            flag_set = False
            db_flags = entry['settings'].get('databaseFlags')
            if db_flags is not None:
                for db_flag in db_flags:
                    if db_flag['name'] == flag and (value is None or db_flag['value'] == value):
                        flag_set = True
                        break
            if flag_set == False:
                details.append(details_msg % (entry['name'], p))
    return details




default_regex_rules = {
    # Private keys
    "RSA private key": "-----BEGIN RSA PRIVATE KEY-----",
    "SSH (DSA) private key": "-----BEGIN DSA PRIVATE KEY-----",
    "SSH (EC) private key": "-----BEGIN EC PRIVATE KEY-----",
    "PGP private key block": "-----BEGIN PGP PRIVATE KEY BLOCK-----",
    # Cloud API keys / Tokens
    "Amazon AWS Access Key ID": "AKIA[0-9A-Z]{16}",
    "Amazon MWS Auth Token": "amzn\\.mws\\.[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
    "AWS API Key": "AKIA[0-9A-Z]{16}",
    "Google API Key": "AIza[0-9A-Za-z\\-_]{35}",
    "Google Cloud Platform API Key": "AIza[0-9A-Za-z\\-_]{35}",
    "Google Cloud Platform OAuth": "[0-9]+-[0-9A-Za-z_]{32}\\.apps\\.googleusercontent\\.com",
    "Google Drive API Key": "AIza[0-9A-Za-z\\-_]{35}",
    "Google Drive OAuth": "[0-9]+-[0-9A-Za-z_]{32}\\.apps\\.googleusercontent\\.com",
    "Google (GCP) Service-account": "\"type\": \"service_account\"",
    "Google Gmail API Key": "AIza[0-9A-Za-z\\-_]{35}",
    "Google Gmail OAuth": "[0-9]+-[0-9A-Za-z_]{32}\\.apps\\.googleusercontent\\.com",
    "Google OAuth Access Token": "ya29\\.[0-9A-Za-z\\-_]+",
    "Google YouTube API Key": "AIza[0-9A-Za-z\\-_]{35}",
    "Google YouTube OAuth": "[0-9]+-[0-9A-Za-z_]{32}\\.apps\\.googleusercontent\\.com",
    "Facebook Access Token": "EAACEdEose0cBA[0-9A-Za-z]+",
    "Facebook OAuth": "[f|F][a|A][c|C][e|E][b|B][o|O][o|O][k|K].*['|\"][0-9a-f]{32}['|\"]",
    "GitHub": "[g|G][i|I][t|T][h|H][u|U][b|B].*['|\"][0-9a-zA-Z]{35,40}['|\"]",
    "Heroku API Key": "[h|H][e|E][r|R][o|O][k|K][u|U].*[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12}",
    "MailChimp API Key": "[0-9a-f]{32}-us[0-9]{1,2}",
    "Mailgun API Key": "key-[0-9a-zA-Z]{32}",
    "Stripe API Key": "sk_live_[0-9a-zA-Z]{24}",
    "Stripe Restricted API Key": "rk_live_[0-9a-zA-Z]{24}",
    "Twitter Access Token": "[t|T][w|W][i|I][t|T][t|T][e|E][r|R].*[1-9][0-9]+-[0-9a-zA-Z]{40}",
    "Twitter OAuth": "[t|T][w|W][i|I][t|T][t|T][e|E][r|R].*['|\"][0-9a-zA-Z]{35,44}['|\"]",
    "Twilio API Key": "SK[0-9a-fA-F]{32}",
    "Square Access Token": "sq0atp-[0-9A-Za-z\\-_]{22}",
    "Square OAuth Secret": "sq0csp-[0-9A-Za-z\\-_]{43}",
    "PayPal Braintree Access Token": "access_token\\$production\\$[0-9a-z]{16}\\$[0-9a-f]{32}",
    "JSON Web Token (JWT)": "eyJ[A-Za-z0-9-_=]+\.[A-Za-z0-9-_=]+\.?[A-Za-z0-9-_.+/=]*?",
    # Misc.
    "Password in URL": "[a-zA-Z]{3,10}://[^/\\s:@]{3,20}:[^/\\s:@]{3,20}@.{1,100}[\"'\\s]",
    "HTTP Authorization Basic": "Authorization:\sBasic\s[a-zA-Z0-9=:_\+\/-]+",
    "HTTP Authorization Bearer": "Authorization:\sBearer\s[a-zA-Z0-9_\-\.=:_\+\/]+"
}

common_passwords = [
"123456",
"password",
"12345678",
"qwerty",
"123456789",
"12345",
"1234",
"111111",
"1234567",
"dragon",
"123123",
"baseball",
"abc123",
"football",
"monkey",
"letmein",
"696969",
"shadow",
"master",
"666666",
"qwertyuiop",
"123321",
"mustang",
"1234567890",
"michael",
"654321",
"pussy",
"superman",
"1qaz2wsx",
"7777777",
"fuckyou",
"121212",
"000000",
"qazwsx",
"123qwe",
"killer",
"trustno1",
"jordan",
"jennifer",
"zxcvbnm",
"asdfgh",
"hunter",
"buster",
"soccer",
"harley",
"batman",
"andrew",
"tigger",
"sunshine",
"iloveyou",
"fuckme",
"2000",
"charlie",
"robert",
"thomas",
"hockey",
"ranger",
"daniel",
"starwars",
"klaster",
"112233",
"george",
"asshole",
"computer",
"michelle",
"jessica",
"pepper",
"1111",
"zxcvbn",
"555555",
"11111111",
"131313",
"freedom",
"777777",
"pass",
"fuck",
"maggie",
"159753",
"aaaaaa",
"ginger",
"princess",
"joshua",
"cheese",
"amanda",
"summer",
"love",
"ashley",
"6969",
"nicole",
"chelsea",
"biteme",
"matthew",
"access",
"yankees",
"987654321",
"dallas",
"austin",
"thunder",
"taylor",
"matrix",
"william",
"corvette",
"hello",
"martin",
"heather",
"secret",
"fucker",
"merlin",
"diamond",
"1234qwer",
"gfhjkm",
"hammer",
"silver",
"222222",
"88888888",
"anthony",
"justin",
"test",
"bailey",
"q1w2e3r4t5",
"patrick",
"internet",
"scooter",
"orange",
"11111",
"golfer",
"cookie",
"richard",
"samantha",
"bigdog",
"guitar",
"jackson",
"whatever",
"mickey",
"chicken",
"sparky",
"snoopy",
"maverick",
"phoenix",
"camaro",
"sexy",
"peanut",
"morgan",
"welcome",
"falcon",
"cowboy",
"ferrari",
"samsung",
"andrea",
"smokey",
"steelers",
"joseph",
"mercedes",
"dakota",
"arsenal",
"eagles",
"melissa",
"boomer",
"booboo",
"spider",
"nascar",
"monster",
"tigers",
"yellow",
"xxxxxx",
"123123123",
"gateway",
"marina",
"diablo",
"bulldog",
"qwer1234",
"compaq",
"purple",
"hardcore",
"banana",
"junior",
"hannah",
"123654",
"porsche",
"lakers",
"iceman",
"money",
"cowboys",
"987654",
"london",
"tennis",
"999999",
"ncc1701",
"coffee",
"scooby",
"0000",
"miller",
"boston",
"q1w2e3r4",
"fuckoff",
"brandon",
"yamaha",
"chester",
"mother",
"forever",
"johnny",
"edward",
"333333",
"oliver",
"redsox",
"player",
"nikita",
"knight",
"fender",
"barney",
"midnight",
"please",
"brandy",
"chicago",
"badboy",
"iwantu",
"slayer",
"rangers",
"charles",
"angel",
"flower",
"bigdaddy",
"rabbit",
"wizard",
"bigdick",
"jasper",
"enter",
"rachel",
"chris",
"steven",
"winner",
"adidas",
"victoria",
"natasha",
"1q2w3e4r",
"jasmine",
"winter",
"prince",
"panties",
"marine",
"ghbdtn",
"fishing",
"cocacola",
"casper",
"james",
"232323",
"raiders",
"888888",
"marlboro",
"gandalf",
"asdfasdf",
"crystal",
"87654321",
"12344321",
"sexsex",
"golden",
"blowme",
"bigtits",
"8675309",
"panther",
"lauren",
"angela",
"bitch",
"spanky",
"thx1138",
"angels",
"madison",
"winston",
"shannon",
"mike",
"toyota",
"blowjob",
"jordan23",
"canada",
"sophie",
"Password",
"apples",
"dick",
"tiger",
"razz",
"123abc",
"pokemon",
"qazxsw",
"55555",
"qwaszx",
"muffin",
"johnson",
"murphy",
"cooper",
"jonathan",
"liverpoo",
"david",
"danielle",
"159357",
"jackie",
"1990",
"123456a",
"789456",
"turtle",
"horny",
"abcd1234",
"scorpion",
"qazwsxedc",
"101010",
"butter",
"carlos",
"password1",
"dennis",
"slipknot",
"qwerty123",
"booger",
"asdf",
"1991",
"black",
"startrek",
"12341234",
"cameron",
"newyork",
"rainbow",
"nathan",
"john",
"1992",
"rocket",
"viking",
"redskins",
"butthead",
"asdfghjkl",
"1212",
"sierra",
"peaches",
"gemini",
"doctor",
"wilson",
"sandra",
"helpme",
"qwertyui",
"victor",
"florida",
"dolphin",
"pookie",
"captain",
"tucker",
"blue",
"liverpool",
"theman",
"bandit",
"dolphins",
"maddog",
"packers",
"jaguar",
"lovers",
"nicholas",
"united",
"tiffany",
"maxwell",
"zzzzzz",
"nirvana",
"jeremy",
"suckit",
"stupid",
"porn",
"monica",
"elephant",
"giants",
"jackass",
"hotdog",
"rosebud",
"success",
"debbie",
"mountain",
"444444",
"xxxxxxxx",
"warrior",
"1q2w3e4r5t",
"q1w2e3",
"123456q",
"albert",
"metallic",
"lucky",
"azerty",
"7777",
"shithead",
"alex",
"bond007",
"alexis",
"1111111",
"samson",
"5150",
"willie",
"scorpio",
"bonnie",
"gators",
"benjamin",
"voodoo",
"driver",
"dexter",
"2112",
"jason",
"calvin",
"freddy",
"212121",
"creative",
"12345a",
"sydney",
"rush2112",
"1989",
"asdfghjk",
"red123",
"bubba",
"4815162342",
"passw0rd",
"trouble",
"gunner",
"happy",
"fucking",
"gordon",
"legend",
"jessie",
"stella",
"qwert",
"eminem",
"arthur",
"apple",
"nissan",
"bullshit",
"bear",
"america",
"1qazxsw2",
"nothing",
"parker",
"4444",
"rebecca",
"qweqwe",
"garfield",
"01012011",
"beavis",
"69696969",
"jack",
"asdasd",
"december",
"2222",
"102030",
"252525",
"11223344",
"magic",
"apollo",
"skippy",
"315475",
"girls",
"kitten",
"golf",
"copper",
"braves",
"shelby",
"godzilla",
"beaver",
"fred",
"tomcat",
"august",
"buddy",
"airborne",
"1993",
"1988",
"lifehack",
"qqqqqq",
"brooklyn",
"animal",
"platinum",
"phantom",
"online",
"xavier",
"darkness",
"blink182",
"power",
"fish",
"green",
"789456123",
"voyager",
"police",
"travis",
"12qwaszx",
"heaven",
"snowball",
"lover",
"abcdef",
"00000",
"pakistan",
"007007",
"walter",
"playboy",
"blazer",
"cricket",
"sniper",
"hooters",
"donkey",
"willow",
"loveme",
"saturn",
"therock",
"redwings"
]



import sys
import os
import subprocess
import logging

from . import utils

docker_cli = ""

def docker_available():
    if os.path.isfile("/usr/bin/docker"):
        return "/usr/bin/docker"
    elif os.path.isfile("/usr/local/bin/docker"):
        return "/usr/local/bin/docker"
    return None 

def start_docker_container(args):
    if args.image is None:
        return None
    cmdarr = [docker_cli+' run -d --rm -i -t '+args.image]
    out = ''
    try:
        out = subprocess.check_output(cmdarr, shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error starting docker container: "+args.image)
        sys.exit(1)
    container_id = out[:12]
    logging.info("Started container with ID ["+container_id+"] from image ["+args.image+"] for discovery")
    return container_id

def stop_docker_container(args, container_id):
    if args.image is None:
        return
    cmdarr = [docker_cli+' stop '+container_id]
    out = ''
    try:
        out = subprocess.check_output(cmdarr, shell=True)
    except subprocess.CalledProcessError:
        logging.error("Error stopping docker container with container ID ["+container_id+"]")
        sys.exit(1)
    logging.info("Stopped container with ID ["+container_id+"]")

def get_os_release(args, container_id):
    base_cmd = docker_cli+' exec -i -t '+container_id+' /bin/sh -c '
    freebsd = False
    out = None
    cmd = '"/bin/cat /etc/os-release"'
    cmdarr = [base_cmd + cmd]
    try:
        out = subprocess.check_output(cmdarr, shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error determining os type for container ID: %s", container_id)

    if out is None or out.strip() == '':
        # try FreeBSD
        cmd = '"/usr/bin/uname -v -p"'
        cmdarr = [base_cmd + cmd]
        try:
            out = subprocess.check_output(cmdarr, shell=True)
            out = out.decode(args.encoding)
        except subprocess.CalledProcessError:
            logging.error("Error determining os type for container ID: %s", container_id)

        if out is not None and 'FreeBSD' not in out:
            # try OpenBSD
            cmd = '"/usr/bin/uname -srvm"'
            cmdarr = [base_cmd + cmd]
            try:
                out = subprocess.check_output(cmdarr, shell=True)
                out = out.decode(args.encoding)
            except subprocess.CalledProcessError:
                logging.error("Error determining os type for container ID: %s", container_id)

    if out is None:
        logging.error("Failed to get os-release")
        return None

    if 'FreeBSD' in out or 'OpenBSD' in out:
        return out
    else:
        output_lines = out.splitlines()
        for l in output_lines:
            if 'PRETTY_NAME' in l:
                return l.split('=')[1].replace('"','')
    return None

def pull_image(args):
    cmdarr = [docker_cli, "pull", args.image]
    try:
        out = subprocess.check_output(cmdarr)
    except subprocess.CalledProcessError:
        logging.error("Error pulling docker image: "+args.image)
        return False
    return True

def get_image_id(args):
    cmdarr = [docker_cli, "images", args.image]
    out = ''
    try:
        out = subprocess.check_output(cmdarr)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error getting image details: "+args.image)
        return None 
    imageid = None
    for l in out.splitlines():
        if 'REPOSITORY' in l:
            continue
        imageid = l.split()[2]
        break
    return imageid

def discover_rh(args, container_id):
    plist = []
    cmdarr = [docker_cli+' exec -i -t '+container_id+' /bin/sh -c "/usr/bin/yum list installed"']
    logging.info("Retrieving product details from image")
    yumout = ''
    try:
        yumout = subprocess.check_output(cmdarr, shell=True)
        yumout = yumout.decode(args.encoding)
    except subprocess.CalledProcessError:
        cmdarr = [docker_cli+' exec -i -t '+container_id+' /bin/sh -c "/usr/bin/rpm -qa"']
        rpmout = ''
        try:
            rpmout = subprocess.check_output(cmdarr, shell=True)
            rpmout = rpmout.decode(args.encoding)
        except subprocess.CalledProcessError:
            logging.error("Error running inventory for container ID [%s]", container_id)
            return None
        for l in rpmout.splitlines():
            tokens = l.split('-')
            length = len(tokens)
            if length <= 2:
                pname = tokens[0]
                version = tokens[1]
            else:
                version = tokens[length-2]+'-'+tokens[length-1]
                pname = "-".join(tokens[:-2])
            logging.debug("Found product [%s %s]", pname, version)
            plist.append(pname+' '+version)
        logging.info("Completed retrieval of product details from image")
        return plist

    begin = False
    for l in yumout.splitlines():
        if 'Installed Packages' in l:
            begin = True
            continue
        if not begin:
            continue
        lsplit = l.split()
        pkg = lsplit[0]
        if len(lsplit) > 1:
            ver = lsplit[1]
        else:
            ver = ''
        pkgsp = pkg.split(".")
        pkg = pkgsp[0]
        arch = pkgsp[1]
        pkg = pkg.replace('\x1b[1m','')
        pkg = pkg.replace('\x1b[1','')
        pkg = pkg.replace('\x1b(B','')
        pkg = pkg.replace('\x1b(m','')
        pkg = pkg.replace('\x1b[m','')
        if ':' in ver:
            ver = ver.split(':')[1]
        ver = ver + "." + arch
        logging.debug("Found product [%s %s]", pkg, ver)
        plist.append(pkg+' '+ver)
    logging.info("Completed retrieval of product details from image")
    return plist

def discover_ubuntu(args, container_id):
    plist = []
    cmdarr = [docker_cli+' exec -i -t '+container_id+' /bin/sh -c "/usr/bin/apt list --installed"']
    logging.info("Retrieving product details from image")
    yumout = ''
    try:
        yumout = subprocess.check_output(cmdarr, shell=True)
        yumout = yumout.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error running inventory for container ID: "+container_id)
        return None 

    begin = False
    for l in yumout.splitlines():
        if 'Listing...' in l:
            begin = True
            continue
        if not begin:
            continue
        if l.strip() == '':
            continue
        lsplit = l.split()
        pkg = lsplit[0].split('/')[0]
        pkg = pkg.replace('\x1b[32m','')
        pkg = pkg.replace('\x1b[0m','')
        ver = lsplit[1]
        logging.debug("Found product [%s %s]", pkg, ver)
        plist.append(pkg+' '+ver)
    logging.info("Completed retrieval of product details from image")
    return plist

def discover_openbsd(args, container_id):
    plist = []
    cmdarr = [docker_cli+' exec -i -t '+container_id+' /bin/sh -c "/usr/sbin/pkg_info -A"']
    logging.info("Retrieving product details from image")
    try:
        pkgout = subprocess.check_output(cmdarr, shell=True)
        pkgout = pkgout.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error running inventory for container ID: %s",container_id)
        return None

    begin = False
    for l in pkgout.splitlines():
        lsplit = l.split()
        pkgline = lsplit[0]
        ldash = pkgline.rfind('-')
        pkg = pkgline[:ldash] + ' ' + pkgline[ldash + 1:]
        logging.debug("Found product [%s]", pkg)
        plist.append(pkg)
    logging.info("Completed retrieval of product details from image")
    return plist

def discover_freebsd(args, container_id):
    plist = []
    cmdarr = [docker_cli+' exec -i -t '+container_id+' /bin/sh -c "/usr/sbin/pkg info"']
    logging.info("Retrieving product details from image")
    try:
        pkgout = subprocess.check_output(cmdarr, shell=True)
        pkgout = pkgout.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error running inventory for container ID: %s",container_id)
        return None

    begin = False
    for l in pkgout.splitlines():
        lsplit = l.split()
        pkgline = lsplit[0]
        ldash = pkgline.rfind('-')
        pkg = pkgline[:ldash] + ' ' + pkgline[ldash + 1:]
        logging.debug("Found product [%s]", pkg)
        plist.append(pkg)
    logging.info("Completed retrieval of product details from image")
    return plist

def discover(args, atype, os_release, container_id):
    handle = args.handle
    token = args.token
    instance = args.instance
    asset_id = None
    if args.assetid == None:
        asset_id = args.image if args.image is not None else args.containerid
    else:
        asset_id = args.assetid
    asset_name = None
    if args.assetname == None:
        asset_name = args.image if args.image is not None else args.containerid
    else:
        asset_name = args.assetname
    asset_id = asset_id.replace('/','-')
    asset_id = asset_id.replace(':','-')
    asset_name = asset_name.replace('/','-')
    asset_name = asset_name.replace(':','-')

    plist = None
    if atype == 'CentOS' or atype == 'Red Hat' or atype == 'Amazon Linux' or atype == 'Oracle Linux':
        plist = discover_rh(args, container_id)
    elif atype == 'Ubuntu' or atype == 'Debian':
        plist = discover_ubuntu(args, container_id)
    elif atype == 'FreeBSD':
        plist = discover_freebsd(args, container_id)
    elif atype == 'OpenBSD':
        plist = discover_openbsd(args, container_id)

    if plist == None or len(plist) == 0:
        logging.error("Could not inventory container ID: "+container_id)
        stop_docker_container(args, container_id)
        sys.exit(1) 

    asset_data = {}
    asset_data['id'] = asset_id
    asset_data['name'] = asset_name
    asset_data['type'] = atype
    asset_data['owner'] = handle
    asset_data['products'] = plist
    asset_tags = []
    asset_tags.append('OS_RELEASE:' + os_release)
    asset_tags.append('Docker')
    asset_tags.append('Container')
    asset_tags.append('Linux')
    asset_tags.append(atype)
    asset_data['tags'] = asset_tags

    return [ asset_data ]

def run_docker_bench(args):
    DBENCH = "/docker-bench-security.sh"

    asset_id = utils.get_ip() if args.assetid is None else args.assetid
    asset_name = asset_id if args.assetname is None else args.assetname
    os_release = utils.get_os_release(args, None)
    if os_release == None:
        logging.error('Unsupported OS type for running docker bench')
        sys.exit(1) 
    atype = utils.get_asset_type(os_release)

    dbench_path = args.docker_bench_home + DBENCH
    if not os.path.isfile(dbench_path) or not os.access(dbench_path, os.X_OK):
        logging.error('Docker bench script not found')
        sys.exit(1) 
    logging.info('Running docker bench script: '+dbench_path)
    try:
        os.chdir(os.path.dirname(args.docker_bench_home))
        out = subprocess.check_output([dbench_path+" 2>/dev/null "], shell=True)
        out = out.decode(args.encoding)
        ansi_escape = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')
        out = ansi_escape.sub('', out)
    except subprocess.CalledProcessError:
        logging.error("Error running docker bench script")
        return None 
    logging.info("docker bench run completed")

    asset_data = {}
    asset_data['id'] = asset_id
    asset_data['name'] = asset_name
    asset_data['type'] = atype
    asset_data['owner'] = args.handle
    asset_data['products'] = [] 
    asset_tags = []
    asset_tags.append('OS_RELEASE:' + os_release)
    asset_tags.append('Docker')
    asset_tags.append('Container')
    asset_tags.append('Linux')
    asset_tags.append(atype)
    asset_data['tags'] = asset_tags 

    findings = []
    details = ''
    issue = {}
    for l in out.splitlines():
        if not l.startswith('[WARN]'):
            continue
        spa = l.split()
        if spa[1] != '*':
            if 'asset_id' in issue:
                issue['details'] = details
                findings.append(issue)
                details = ''
                issue = {}
            issue['twc_id'] = 'docker-bench-check-'+spa[1].strip()
            issue['asset_id'] = asset_id 
            issue['twc_title'] = l.split('-')[1].strip()
            issue['rating'] = '4'
            issue['object_id'] = '' 
            issue['object_meta'] = ''
            details = ''
        else:
            details = details + l.split('*')[1] + '\n'
    # add the final issue
    if 'asset_id' in issue:
        issue['details'] = details
        findings.append(issue)
    asset_data['config_issues'] = findings
    # disable scan
    args.no_scan = True
    return [ asset_data ]

def get_inventory(args):
    global docker_cli

    if os.geteuid() != 0:
        logging.error("Docker operations need root privilege. Please run as 'sudo' or 'root'")
        sys.exit(1)

    docker_cli = docker_available()
    if not docker_cli:
        sys.exit(1)

    if args.image is None and args.containerid is None:
        logging.error("Error either docker image (--image) or running container id (--containerid) parameter needs to be specified")
        sys.exit(1)

    if args.image is not None:
        if not get_image_id(args):
            if not pull_image(args):
                sys.exit(1)
        container_id = start_docker_container(args)
    elif args.containerid is not None:
        container_id = args.containerid

    os_release = get_os_release(args, container_id)
    if os_release is None:
        stop_docker_container(args, container_id)
        sys.exit(1)
    atype = utils.get_asset_type(os_release)
    if atype is None:
        stop_docker_container(args, container_id)
        sys.exit(1)

    assets = discover(args, atype, os_release, container_id)
    stop_docker_container(args, container_id)
    return assets



import sys
import json
import logging
from . import gcp_cis_utils as gcp_cis_utils

def check7_1():
    # 7.1 Ensure that BigQuery datasets are not anonymously or publicly accessible (Scored)

    logging.info("7.1 Ensure that BigQuery datasets are not anonymously or publicly accessible (Scored)")
    details = []
    projects = gcp_cis_utils.get_bigquery_enabled_projects()
    for p in projects:
        output = gcp_cis_utils.run_cmd("bq ls --project_id=%s --format=json 2>/dev/null" % p)
        try:
            out_json = json.loads(output)
        except ValueError:
            logging.warn("Unable to load response JSON for project [%s]...skipping it", p)
            out_json = []
        for entry in out_json:
            if entry['kind'] != "bigquery#dataset":
                continue
            dataset_id = entry['id']
            output = gcp_cis_utils.run_cmd("bq show --format=json %s 2>/dev/null" %dataset_id)
            try:
                out_json_2 = json.loads(output) 
            except ValueError:
                logging.warn("Unable to load response JSON for project [%s]...skipping it", p)
                out_json_2 = {}
            access_configs = out_json_2.get('access')
            if access_configs is not None:
                for ac in access_configs:
                    if ac.get('specialGroup') == "allAuthenticatedUsers" or ac.get('iamMember') == "allUsers":
                        details.append("BigQuery dataset [%s] in project [%s] is anonymously or publicly accessible" % (dataset_id.split(':')[1], p))
                        break
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-7.1', '7.1 [Level 1] Ensure that BigQuery datasets are not anonymously or publicly accessible (Scored)', "\n".join(details), '4', '', '')
    return None

def run_checks():
    config_issues = []
    gcp_cis_utils.append_issue(config_issues, check7_1())
    return config_issues




# MIT License
# Copyright (c) 2018 Mathieu Durand
# @Author Mathieu Durand
# Version 1.1

import subprocess
import sys
import datetime
import os
import logging
import tempfile
from html.parser import HTMLParser

from . import az_cis_check1 as check1
from . import az_cis_check2 as check2
from . import az_cis_check3 as check3
from . import az_cis_check4 as check4
from . import az_cis_check5 as check5
from . import az_cis_check6 as check6
from . import az_cis_check7 as check7
from . import az_cis_check8 as check8
from . import az_cis_utils as az_cis_utils

################ CSV EXPORT HELPERS ###################
# Use HTML Parser
class MLStripper(HTMLParser):
    def __init__(self):
        self.reset()
        self.strict = False
        self.convert_charrefs= True
        self.fed = []
    def handle_data(self, d):
        self.fed.append(d)
    def get_data(self):
        return ''.join(self.fed)

# Strip HTML tags
def strip_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()

# Process a CSV Column: strip HTML tags, wrap with double-quotes, replace double-quotes with single-quotes, replace line breaks with newline
def c(text):
    #return '"'+ str(text).strip().replace("<br>", "\n").replace("<b>","'").replace("</b>","'").replace('"',"'") + '",'
    return '"'+ str(text).strip().replace("<br>", "\n").replace('"',"'") + '",'

# Process a CSV Row: Add Subscription name column, add newline
subname=""
def r(text):
    return '"' + subname + '",' + text + "\n"

################ HTML HEADER ###################
def generate_report(subid,name,cloudname):
    global subname
    subname=name

    html_start = """
<html>
<head>
<title>Azure CIS Benchmark</title>
</head>
<body>
<h1><center>Azure CIS Benchmark for """+name+"""</center></h1>
<font size="-1">
"""
    html_end = """
</body>
</html>
"""

    start_list="<li><b>\n"
    end_list="</h3></li></b>\n"

################ HTML 1.x ##############
    content1 = """
<h2>CIS Azure 1.x</h2>
<ul>
"""
    content11="1.1 Ensure that multi-factor authentication is enabled for all privileged users"
    content12="1.2 Ensure that multi-factor authentication is enabled for all nonprivileged users"
    content13="1.3 Ensure that there are no guest users"
    content14="1.4 Ensure that 'Allow users to remember multi-factor authentication on devices they trust' is 'Disabled'"
    content15="1.5 Ensure that 'Number of methods required to reset' is set to '2'"
    content16="1.6 Ensure that 'Number of days before users are asked to re-confirm their authentication information' is not set to '0"
    content17="1.7 Ensure that 'Notify users on password resets?' is set to 'Yes'"
    content18="1.8 Ensure that 'Notify all admins when other admins reset their password?' is set to 'Yes'"
    content19="1.9 Ensure that 'Users can consent to apps accessing company data on their behalf' is set to 'No'"
    content110="1.10 Ensure that 'Users can add gallery apps to their Access Panel' is set to 'No'"
    content111="1.11 Ensure that 'Users can register applications' is set to 'No'"
    content112="1.12 Ensure that 'Guest users permissions are limited' is set to 'Yes'"
    content113="1.13 Ensure that 'Members can invite' is set to 'No'"
    content114="1.14 Ensure that 'Guests can invite' is set to 'No'"
    content115="1.15 Ensure that 'Restrict access to Azure AD administration portal' is set to 'Yes'"
    content116="1.16 Ensure that 'Self-service group management enabled' is set to 'No'"
    content117="1.17 Ensure that 'Users can create security groups' is set to 'No'"
    content118="1.18 Ensure that 'Users who can manage security groups' is set to 'None'"
    content119="1.19 Ensure that 'Users can create Office 365 groups' is set to 'No'"
    content120="1.20 Ensure that 'Users who can manage Office 365 groups' is set to 'None'"
    content121="1.21 Ensure that 'Enable 'All Users' group' is set to 'Yes'"
    content122="1.22 Ensure that 'Require Multi-Factor Auth to join devices' is set to 'Yes'"
    content123="1.23 Ensure that no custom subscription owner roles are created"

    result11=check1.check11(subid)
    #sys.exit(0)
    result12=check1.check12()
    result13=check1.check13()
    result14=check1.check14()
    result15=check1.check15()
    result16=check1.check16()
    result17=check1.check17()
    result18=check1.check18()
    result19=check1.check19()
    result110=check1.check110()
    result111=check1.check111()
    result112=check1.check112()
    result113=check1.check113()
    result114=check1.check114()
    result115=check1.check115()
    result116=check1.check116()
    result117=check1.check117()
    result118=check1.check118()
    result119=check1.check119()
    result120=check1.check120()
    result121=check1.check121()
    result122=check1.check122()
    result123=check1.check123()
    content1_1 = '<h3 id="content11">'+start_list+content11+end_list+result11[0]+'<h3 id="content12">'+start_list+content12+end_list+result12[0]+'<h3 id="content13">'+start_list+content13+end_list+result13[0]
    content1_2 = '<h3 id="content11">'+start_list+content14+end_list+result14[0]+'<h3 id="content15">'+start_list+content15+end_list+result15[0]+'<h3 id="content16">'+start_list+content16+end_list+result16[0]
    content1_3 = '<h3 id="content17">'+start_list+content17+end_list+result17[0]+'<h3 id="content18">'+start_list+content18+end_list+result18[0]+'<h3 id="content19">'+start_list+content19+end_list+result19[0]
    content1_4 = '<h3 id="content110">'+start_list+content110+end_list+result110[0]+'<h3 id="content111">'+start_list+content111+end_list+result111[0]+'<h3 id="content112">'+start_list+content112+end_list+result112[0]
    content1_5 = '<h3 id="content113">'+start_list+content113+end_list+result113[0]+'<h3 id="content114">'+start_list+content114+end_list+result114[0]+'<h3 id="content115">'+start_list+content115+end_list+result115[0]
    content1_6 = '<h3 id="content116">'+start_list+content116+end_list+result116[0]+'<h3 id="content117">'+start_list+content117+end_list+result117[0]+'<h3 id="content118">'+start_list+content118+end_list+result118[0]
    content1_7 = '<h3 id="content119">'+start_list+content119+end_list+result119[0]+'<h3 id="content120">'+start_list+content120+end_list+result120[0]+'<h3 id="content121">'+start_list+content121+end_list+result121[0]
    content1_8 = '<h3 id="content122">'+start_list+content122+end_list+result122[0]+'<h3 id="content123">'+start_list+content123+end_list+result123[0]

    content1=content1+content1_1+content1_2+content1_3+content1_4+content1_5+content1_6+content1_7+content1_8 


################ HTML 2.x ##############
    content2 = """
<h2>CIS Azure 2.x</h2>
<ul>
"""
    content21="2.1 Ensure that standard pricing tier is selected"
    content22="2.2 Ensure that 'Automatic provisioning of monitoring agent' is set to 'On'"
    content23="2.3 Ensure that 'System updates' is set to 'On'"
    content24="2.4 Ensure that 'Security Configurations' is set to 'On'"
    content25="2.5 Ensure that 'Endpoint protection' is set to 'On'"
    content26="2.6 Ensure that 'Disk encryption' is set to 'On'"
    content27="2.7 Ensure that 'Network security groups' is set to 'On'"
    content28="2.8 Ensure that 'Web application firewall' is set to 'On'"
    content29="2.9 Ensure that 'Next generation firewall' is set to 'On'"
    content210="2.10 Ensure that 'Vulnerability assessment' is set to 'On'"
    content211="2.11 Ensure that 'Storage Encryption' is set to 'On'"
    content212="2.12 Ensure that 'JIT Network Access' is set to 'On'"
    content213="2.13 Ensure that 'Adaptive Application Controls' is set to 'On'"
    content214="2.14 Ensure that 'SQL auditing & Threat detection' is set to 'On'"
    content215="2.15 Ensure that 'SQL Encryption' is set to 'On'"
    content216="2.16 Ensure that 'Security contact emails' is set"
    content217="2.17 Ensure that security contact 'Phone number' is set"
    content218="2.18 Ensure that 'Send me emails about alerts' is set to 'On'"
    content219="2.19 Ensure that 'Send email also to subscription owners' is set to 'On'"

    result21=check2.check21()
    result22=check2.check22(subid)
    content2_1 = '<h3 id="content21">'+start_list+content21+end_list+result21+'<h3 id="content22">'+start_list+content22+end_list+result22[0]+'<h3 id="content23">'+start_list+content23+end_list+result22[1]
    content2_2 = '<h3 id="content24">'+start_list+content24+end_list+result22[2]+'<h3 id="content25">'+start_list+content25+end_list+result22[3]+'<h3 id="content26">'+start_list+content26+end_list+result22[4]+'<h3 id="content27">'+start_list+content27+end_list+result22[5]+'<h3 id="content28">'+start_list+content28+end_list+result22[6]
    content2_3 = '<h3 id="content29">'+start_list+content29+end_list+result22[7]+'<h3 id="content210">'+start_list+content210+end_list+result22[8]+'<h3 id="content211">'+start_list+content211+end_list+result22[9]
    content2_4 = '<h3 id="content212">'+start_list+content212+end_list+result22[10]+'<h3 id="content213">'+start_list+content213+end_list+result22[11]+'<h3 id="content214">'+start_list+content214+end_list+result22[12]
    content2_5 = '<h3 id="content215">'+start_list+content215+end_list+result22[13]+'<h3 id="content216">'+start_list+content216+end_list+result22[14]+'<h3 id="content217">'+start_list+content217+end_list+result22[15]
    content2_6 = '<h3 id="content218">'+start_list+content218+end_list+result22[16]+'<h3 id="content219">'+start_list+content219+end_list+result22[17]
    content2 = content2 + content2_1 + content2_2 + content2_3 + content2_4 + content2_5+ content2_6

    ############ HTML 3.x ##############
    content3 = """
<h2>CIS Azure 3.x</h2>
<ul>
"""
    result31=check3.check31()
    result32=check3.check32()
    result33=check3.check33()
    result34=check3.check34()
    result35=check3.check35()
    result36=check3.check36()
    result37=check3.check37()
    content31="3.1 Ensure that 'Secure transfer required' is set to 'Enabled'"
    content32="3.2 Ensure that 'Storage service encryption' is set to Enabled for Blob Service"
    content33="3.3 Ensure that storage account access keys are periodically regenerated"
    content34="3.4 Ensure that shared access signature tokens expire within an hour"
    content35="3.5 Ensure that shared access signature tokens are allowed only over https"
    content36="3.6 Ensure that 'Storage service encryption' is set to Enabled for File Service"
    content37="3.7 Ensure that 'Public access level' is set to Private for blob containers"
    

    content3_1 = '<h3 id="content31">'+start_list+content31+end_list+result31[0]
    content3_2 = '<h3 id="content32">'+start_list+content32+end_list+result32[0]
    content3_3 = '<h3 id="content33">'+start_list+content33+end_list+result33[0]
    content3_4 = '<h3 id="content34">'+start_list+content34+end_list+result34+'<h3 id="content35">'+start_list+content35+end_list+result35+'<h3 id="content36">'+start_list+content36+end_list+result36[0]+'<h3 id="content37">'+start_list+content37+end_list+result37[0]
    content3 = content3+content3_1+content3_2+content3_3+content3_4

################ HTML 4.x ##############
    content4 = """
<h2>CIS Azure 4.x</h2>
<ul>
"""
    content411="4.1.1 Ensure that 'Auditing' is set to 'On'"
    content412="4.1.2 Ensure that 'Threat Detection' is set to 'On'"
    content413="4.1.3 Ensure that 'Threat Detection types' is set to 'All'"
    content414="4.1.4 Ensure that 'Send alerts to' is set"
    content415="4.1.5 Ensure that 'Email service and co-administrators' is 'Enabled'"
    content416="4.1.6 Ensure that 'Auditing' Retention is 'greater than 90 days'"
    content417="4.1.7 Ensure that 'Threat Detection' Retention is 'greater than 90 days'"
    content418="4.1.8 Ensure that Azure Active Directory Admin is configured"

    result41 = check4.check41(subid)
    content41_1 = start_list+content411+end_list#+check41()+start_list+content412+end_list+check41()+start_list+content413+end_list+check41()+end_list
    content41_2 = start_list+content414+end_list#+check41()+start_list+content415+end_list+check41()+start_list+content416+end_list+check41()+start_list+content417+end_list+check41()+start_list+content418+end_list+check41()
    content41 = content41_1+content41_2

    
    content421="4.2.1 Ensure that 'Auditing' is set to 'On'"
    content422="4.2.2 Ensure that 'Threat Detection' is set to 'On'"
    content423="4.2.3 Ensure that 'Threat Detection types' is set to 'All'"
    content424="4.2.4 Ensure that 'Send alerts to' is set"
    content425="4.2.5 Ensure that 'Email service and co-administrators' is 'Enabled'"
    content426="4.2.6 Ensure that 'Data encryption' is set to 'On'"
    content427="4.2.7 Ensure that 'Auditing' Retention is 'greater than 90 days'"
    content428="4.2.8 Ensure that 'Threat' Retention is 'greater than 90 days'"

    result42 = check4.check42(subid)

    content42_1 = '<h3 id="content421">'+start_list+content421+end_list+result42[0][0]+'<h3 id="content422">'+start_list+content422+end_list+result42[1][0]+'<h3 id="content423">'+start_list+content423+end_list+result42[2][0]
    content42_2 = '<h3 id="content424">'+start_list+content424+end_list+result42[3][0]+'<h3 id="content425">'+start_list+content425+end_list+result42[4][0]+'<h3 id="content426">'+start_list+content426+end_list+result42[5][0]
    content42_3 = '<h3 id="content427">'+start_list+content427+end_list+result42[6][0]+'<h3 id="content428">'+start_list+content428+end_list+result42[7][0]
    content42 = content42_1+content42_2+content42_3

    content4 = content4 + content41 + content42


################ HTML 5.x ##############
    content5 = """
<h2>CIS Azure 5.x</h2>
<ul>
"""
    content51="5.1 Ensure that a Log Profile exists"
    content52="5.2 Ensure that Activity Log Retention is set 365 days or greater"
    content53="5.3 Ensure that Activity Log Alert exists for Create Policy Assignment"
    content54="5.4 Ensure that Activity Log Alert exists for Create or Update Network Security Group"
    content55="5.5 Ensure that Activity Log Alert exists for Delete Network Security Group"
    content56="5.6 Ensure that Activity Log Alert exists for Create or Update Network Security Group Rule"
    content57="5.7 Ensure that Activity Log Alert exists for Delete Network Security Group Rule"
    content58="5.8 Ensure that Activity Log Alert exists for Create or Update Security Solution"
    content59="5.9 Ensure that Activity Log Alert exists for Delete Security Solution"
    content510="5.10 Ensure that Activity Log Alert exists for Create or Update SQL Server Firewall Rule"
    content511="5.11 Ensure that Activity Log Alert exists for Delete SQL Server Firewall Rule"
    content512="5.12 Ensure that Activity Log Alert exists for Update Security Policy"
    content513="5.13 Ensure that logging for Azure KeyVault is 'Enabled'"

    result5=check5.check50(subid)

    content5_1 = '<h3 id="content51">'+start_list+content51+end_list+result5[0]+'<h3 id="content52">'+start_list+content52+end_list+result5[1]+'<h3 id="content53">'+start_list+content53+end_list+result5[2]
    content5_2 = '<h3 id="content54">'+start_list+content54+end_list+result5[3]+'<h3 id="content55">'+start_list+content55+end_list+result5[4]+'<h3 id="content56">'+start_list+content56+end_list+result5[5]+'<h3 id="content57">'+start_list+content57+end_list+result5[6]+'<h3 id="content58">'+start_list+content58+end_list+result5[7]
    content5_3 = '<h3 id="content59">'+start_list+content59+end_list+result5[8]+'<h3 id="content510">'+start_list+content510+end_list+result5[9]+'<h3 id="content511">'+start_list+content511+end_list+result5[10]
    content5_4 = '<h3 id="content512">'+start_list+content512+end_list+result5[11]+'<h3 id="content513">'+start_list+content513+end_list+result5[12]
    content50 = content5_1 + content5_2 + content5_3 + content5_4
    content5 = content5 + content50


################ HTML 6.x ##############
    content6 = """
<h2>CIS Azure 6.x</h2>
<ul>
"""
    result62=check6.check62(subid)
    result63=check6.check63()
    result64=check6.check64(subid)
    result65=check6.check65(subid)
    content61="6.1 Ensure that RDP access is restricted from the internet"
    content62="6.2 Ensure that SSH access is restricted from the internet"
    content63="6.3 Ensure that SQL server access is restricted from the internet"
    content64="6.4 Ensure that Network Security Group Flow Log retention period is greater than 90 days"
    content65="6.5 Ensure that Network Watcher is Enabled"
    
    content6_1 = '<h3 id="content61">'+start_list+content61+end_list+result62[0][0]+'<h3 id="content62">'+start_list+content62+end_list+result62[1][0]+'<h3 id="content63">'+start_list+content63+end_list+result63[0]
    content6_2 = '<h3 id="content64">'+start_list+content64+end_list+result64[0]
    content6_3 = '<h3 id="content65">'+start_list+content65+end_list+result65[0]

    content6=content6+content6_1+content6_2+content6_3

################ HTML 7.x ##############
    content7 = """
<h2>CIS Azure 7.x</h2>
<ul>
"""

    content71="7.1 Ensure that VM agent is installed"
    content72="7.2 Ensure that 'OS disk' are encrypted"
    content73="7.3 Ensure that 'Data disks' are encrypted"
    content74="7.4 Ensure that only approved extensions are installed"
    content75="7.5 Ensure that the latest OS Patches for all Virtual Machines are applied"
    content76="7.6 Ensure that the endpoint protection for all Virtual Machines is installed"

    result7=check7.check70(subid)

    content7_1 = '<h3 id="content71">'+start_list+content71+end_list+result7[0]+'<h3 id="content72">'+start_list+content72+end_list+result7[1]+'<h3 id="content73">'+start_list+content73+end_list+result7[2]
    content7_2 = '<h3 id="content74">'+start_list+content74+end_list+result7[3]+'<h3 id="content75">'+start_list+content75+end_list+result7[4]+'<h3 id="content76">'+start_list+content76+end_list+result7[5]
    content7=content7+content7_1+content7_2


################ HTML 8.x ##############
    content8 = """
<h2>CIS Azure 8.x</h2>
<ul>
"""
    
    content81="8.1 Ensure that the expiry date is set on all Keys"
    content82="8.2 Ensure that the expiry date is set on all Secrets"
    content83="8.3 Ensure that Resource Locks are set for mission critical Azure resources"
    
    result80=check8.check80()
    result83=check8.check83()
    content8 = content8+'<h3 id="content81">'+start_list+content81+end_list+result80[0][0]+'<h3 id="content82">'+start_list+content82+end_list+result80[1][0]+'<h3 id="content83">'+start_list+content83+end_list+result83
    
    logging.info("Finished Queries")
    logging.info("Generating Report")
############## Summary Table ##########
    summary0 = """
<h1 id="Summary">1 - Score Summary</h1>
<style>
table, th, td {
    border: 1px solid black;
    border-collapse: collapse;
}
th, td {
    padding: 5px;
    font-size: small;
th {
    text-align: left;
}
</style>
<table style="width:80%">
<tr>
    <th>Check Description</th> 
    <th>Tests</th>
    <th>Score</th>
    <th>Comments</th>
</tr>
"""
########################################################################################################################
############################################################ Summary 1 #################################################
########################################################################################################################
#Calculate % Passed

    perc13=round(100*result13[1]/result13[2],2)
    perc123=round(100*result123[1]/result123[2],2)

    calc1=(perc13+perc123)/2
    score1= round(calc1,2)

    summary1 = """
<tr>
    <td><b>1 Identity and Access Management</b></td> 
    <td></td>
    <td><b>"""+str(score1)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content14">"""+content11+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content14">"""+content12+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content13">"""+content13+"""</a></td> 
    <td>"""+result13[3]+str(result13[1])+"""/"""+str(result13[2])+"""</td>
    <td>"""+str(perc13)+"""%</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content14">"""+content14+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content15">"""+content15+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content16">"""+content16+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content17">"""+content17+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content18">"""+content18+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content19">"""+content19+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content110">"""+content110+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content111">"""+content111+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content112">"""+content112+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content113">"""+content113+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content114">"""+content114+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content115">"""+content115+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content116">"""+content116+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content117">"""+content117+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content118">"""+content118+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content119">"""+content119+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content120">"""+content120+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content121">"""+content121+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content122">"""+content122+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Active Directory Check</td>
</tr>
<tr>
    <td><a href="#content123">"""+content123+"""</a></td> 
    <td>"""+result123[3]+str(result123[1])+"""/"""+str(result123[2])+"""</td>
    <td>"""+str(perc123)+"""%</td>
    <td></td>
</tr>
"""

########################################################################################################################
############################################################ Summary 2 #################################################
########################################################################################################################
#Calculate % Passed, check 21 is ignored

    calc2=100*(result22[18][1]+result22[19][1]+result22[20][1]+result22[21][1]+result22[22][1]+result22[23][1]+result22[24][1]+result22[25][1]+result22[26][1]+result22[27][1]+result22[28][1]+result22[29][1]+result22[30][1]+result22[31][1]+result22[32][1]+result22[33][1]+result22[34][1]+result22[35][1])/18
    
    score2= round(calc2,2)

    summary2 = """
<tr>
    <td><b>2 Security Center</b></td> 
    <td></td>
    <td><b>"""+str(score2)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content21">"""+content21+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Standard pricing 15.00 USD/NODE/MONTH</td>
</tr>
<tr>
    <td><a href="#content22">"""+content22+"""</a></td> 
    <td>"""+result22[18][0]+"""</td>
    <td>"""+str(100*result22[18][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content23">"""+content23+"""</a></td> 
    <td>"""+result22[19][0]+"""</td>
    <td>"""+str(100*result22[19][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content24">"""+content24+"""</a></td> 
    <td>"""+result22[20][0]+"""</td>
    <td>"""+str(100*result22[20][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content25">"""+content25+"""</a></td> 
    <td>"""+result22[21][0]+"""</td>
    <td>"""+str(100*result22[21][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content26">"""+content26+"""</a></td> 
    <td>"""+result22[22][0]+"""</td>
    <td>"""+str(100*result22[22][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content27">"""+content27+"""</a></td> 
    <td>"""+result22[23][0]+"""</td>
    <td>"""+str(100*result22[23][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content28">"""+content28+"""</a></td> 
    <td>"""+result22[24][0]+"""</td>
    <td>"""+str(100*result22[24][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content29">"""+content29+"""</a></td> 
    <td>"""+result22[25][0]+"""</td>
    <td>"""+str(100*result22[25][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content210">"""+content210+"""</a></td> 
    <td>"""+result22[26][0]+"""</td>
    <td>"""+str(100*result22[26][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content211">"""+content211+"""</a></td> 
    <td>"""+result22[27][0]+"""</td>
    <td>"""+str(100*result22[27][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content212">"""+content212+"""</a></td> 
    <td>"""+result22[28][0]+"""</td>
    <td>"""+str(100*result22[28][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content213">"""+content213+"""</a></td> 
    <td>"""+result22[29][0]+"""</td>
    <td>"""+str(100*result22[29][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content214">"""+content214+"""</a></td> 
    <td>"""+result22[30][0]+"""</td>
    <td>"""+str(100*result22[30][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content215">"""+content215+"""</a></td> 
    <td>"""+result22[31][0]+"""</td>
    <td>"""+str(100*result22[31][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content216">"""+content216+"""</a></td> 
    <td>"""+result22[32][0]+"""</td>
    <td>"""+str(100*result22[32][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content217">"""+content217+"""</a></td> 
    <td>"""+result22[33][0]+"""</td>
    <td>"""+str(100*result22[33][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content218">"""+content218+"""</a></td> 
    <td>"""+result22[34][0]+"""</td>
    <td>"""+str(100*result22[34][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content219">"""+content219+"""</a></td> 
    <td>"""+result22[35][0]+"""</td>
    <td>"""+str(100*result22[35][1])+"""%</td>
    <td></td>
</tr>
"""
    logging.info("Finished Summary 2")

########################################################################################################################
############################################################ Summary 3 #################################################
########################################################################################################################
#Calculate % Passed, 33,34 not scored and 35 ignored

    perc31=round(100*result31[1]/result31[2],2)
    perc32=round(100*result32[1]/result32[2],2)
    #perc33=round(100*result33[1]/result33[2],2)
    perc36=round(100*result36[1]/result36[2],2)
    perc37=round(100*result37[1]/result37[2],2)

    calc3=(perc31+perc32+perc36+perc37)/4
    score3= round(calc3,2)

    summary3 = """
<tr>
    <td><b>3 Storage Accounts</b></td> 
    <td></td>
    <td><b>"""+str(score3)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content31">"""+content31+"""</a></td> 
    <td>"""+result31[3]+str(result31[1])+"""/"""+str(result31[2])+"""</td>
    <td>"""+str(perc31)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content32">"""+content32+"""</a></td> 
    <td>"""+result32[3]+str(result32[1])+"""/"""+str(result32[2])+"""</td>
    <td>"""+str(perc32)+"""%</td>
    <td>By default, data is encrypted using Microsoft Managed Keys for Azure Blobs, Tables, Files and Queues.</td>
</tr>
<tr>
    <td><a href="#content33">"""+content33+"""</a></td> 
    <td>"""+result33[3]+str(result33[1])+"""/"""+str(result33[2])+"""</td>
    <td>Not Scored</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content34">"""+content34+"""</a></td> 
    <td>Not Scored</td>
    <td>Not Scored</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content35">"""+content35+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content36">"""+content36+"""</a></td> 
    <td>"""+result36[3]+str(result36[1])+"""/"""+str(result36[2])+"""</td>
    <td>"""+str(perc36)+"""%</td>
    <td>By default, data is encrypted using Microsoft Managed Keys for Azure Blobs, Tables, Files and Queues.</td>
</tr>
<tr>
    <td><a href="#content37">"""+content37+"""</a></td> 
    <td>"""+result37[3]+str(result37[1])+"""/"""+str(result37[2])+"""</td>
    <td>"""+str(perc37)+"""%</td>
    <td></td>
</tr>
"""

    logging.info("Finished Summary3")

########################################################################################################################
############################################################ Summary 4 #################################################
########################################################################################################################
#Calculate % Passed

    if (len(result41)>1):
        logging.info("To be done")
        score41 = 0
        summary41 = """
<tr>
    <td><b>4.1 SQL Servers</b></td> 
    <td></td>
    <td><b>"""+str(score41)+"""%</b></td> 
</tr>
"""
    else:
        score41 = 100
        summary41 = """
<tr>
    <td><b>4.1 SQL Servers</b></td> 
    <td></td>
    <td><b></b></td> 
    <td></td>
</tr>
<tr>
    <td><a href="#content411">"""+content411+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Also under Security Center - Resource security hygiene</td>
</tr>
<tr>
    <td><a href="#content412">"""+content412+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td>Also under Security Center - Resource security hygiene</td>
</tr>
<tr>
    <td><a href="#content413">"""+content413+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content414">"""+content414+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content415">"""+content415+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content416">"""+content416+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content417">"""+content417+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content418">"""+content418+"""</a></td> 
    <td>Unavailable</td>
    <td>Unavailable</td>
    <td></td>
</tr>
"""

    #if (len(result42)>1):
    calc42=100*(result42[0][1]+result42[0][1]+result42[1][1]+result42[2][1]+result42[3][1]+result42[4][1]+result42[5][1]+result42[6][1]+result42[7][1])/8

    perc421=round(100*result42[0][1]/result42[0][2],2)
    perc422=round(100*result42[1][1]/result42[1][2],2)
    perc423=round(100*result42[2][1]/result42[2][2],2)
    perc424=round(100*result42[3][1]/result42[3][2],2)
    perc425=round(100*result42[4][1]/result42[4][2],2)
    perc426=round(100*result42[5][1]/result42[5][2],2)
    perc427=round(100*result42[6][1]/result42[6][2],2)
    perc428=round(100*result42[7][1]/result42[7][2],2)

    calc42=(perc421+perc422)/2
    score42 = round(calc42,2)
    score4= (score41+score42)/2
    summary42 = """
<tr>
    <td><b>4.2 SQL Databases</b></td> 
    <td></td>
    <td><b>"""+str(score4)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content421">"""+content421+"""</a></td> 
    <td>"""+result42[0][3]+str(result42[0][1])+"""/"""+str(result42[0][2])+"""</td>
    <td>"""+str(perc421)+"""%</td>
    <td>Also under Security Center - Resource security hygiene</td>
</tr>
<tr>
    <td><a href="#content422">"""+content422+"""</a></td> 
    <td>"""+result42[1][3]+str(result42[1][1])+"""/"""+str(result42[1][2])+"""</td>
    <td>"""+str(perc422)+"""%</td>
    <td>Also under Security Center - Resource security hygiene. Turn on Advanced Threat Protection for all databases on this server, at the cost of 15 USD/server/month</td>
</tr>
<tr>
    <td><a href="#content423">"""+content423+"""</a></td> 
    <td>"""+result42[2][3]+str(result42[2][1])+"""/"""+str(result42[2][2])+"""</td>
    <td>"""+str(perc423)+"""%</td>
    <td>Turn on Advanced Threat Protection for all databases on this server, at the cost of 15 USD/server/month</td>
</tr>
<tr>
    <td><a href="#content424">"""+content424+"""</a></td> 
    <td>"""+result42[3][3]+str(result42[3][1])+"""/"""+str(result42[3][2])+"""</td>
    <td>"""+str(perc424)+"""%</td>
    <td>Turn on Advanced Threat Protection for all databases on this server, at the cost of 15 USD/server/month</td>
</tr>
<tr>
    <td><a href="#content425">"""+content425+"""</a></td> 
    <td>"""+result42[4][3]+str(result42[4][1])+"""/"""+str(result42[4][2])+"""</td>
    <td>"""+str(perc425)+"""%</td>
    <td>Turn on Advanced Threat Protection for all databases on this server, at the cost of 15 USD/server/month</td>
</tr>
<tr>
    <td><a href="#content426">"""+content426+"""</a></td> 
    <td>"""+result42[5][3]+str(result42[5][1])+"""/"""+str(result42[5][2])+"""</td>
    <td>"""+str(perc426)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content427">"""+content427+"""</a></td> 
    <td>"""+result42[6][3]+str(result42[6][1])+"""/"""+str(result42[6][2])+"""</td>
    <td>"""+str(perc427)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content428">"""+content428+"""</a></td> 
    <td>"""+result42[7][3]+str(result42[7][1])+"""/"""+str(result42[7][2])+"""</td>
    <td>"""+str(perc428)+"""%</td>
    <td>Turn on Advanced Threat Protection for all databases on this server, at the cost of 15 USD/server/month</td>
</tr>
"""

    score4= (score41+score42)/2
    summary4 = """
<tr>
    <td><b>4 SQL Services</b></td> 
    <td></td>
    <td><b>"""+str(score4)+"""%</b></td> 
</tr>
"""
    summary4=summary4+summary41+summary42
    logging.info("Finished Summary 4")

########################################################################################################################
############################################################ Summary 5 #################################################
########################################################################################################################
#Calculate % Passed, 51 and 52 ignored

    calc5=100*(result5[13][1]+result5[14][1]+result5[15][1]+result5[16][1]+result5[17][1]+result5[18][1]+result5[19][1]+result5[20][1]+result5[21][1]+result5[22][1]+result5[23][1]+result5[24][1]+result5[22][1])/13
    score5= round(calc5,2)

    summary5 = """
<tr>
    <td><b>5 Logging and Monitoring</b></td> 
    <td></td>
    <td><b>"""+str(score5)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content51">"""+content51+"""</a></td> 
    <td>"""+result5[13][0]+"""</td>
    <td>"""+str(100*result5[13][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content52">"""+content52+"""</a></td> 
    <td>"""+result5[14][0]+"""</td>
    <td>"""+str(100*result5[14][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content53">"""+content53+"""</a></td> 
    <td>"""+result5[15][0]+"""</td>
    <td>"""+str(100*result5[15][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content54">"""+content54+"""</a></td> 
    <td>"""+result5[16][0]+"""</td>
    <td>"""+str(100*result5[16][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content55">"""+content55+"""</a></td> 
    <td>"""+result5[17][0]+"""</td>
    <td>"""+str(100*result5[17][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content56">"""+content56+"""</a></td> 
    <td>"""+result5[18][0]+"""</td>
    <td>"""+str(100*result5[18][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content57">"""+content57+"""</a></td> 
    <td>"""+result5[19][0]+"""</td>
    <td>"""+str(100*result5[19][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content58">"""+content58+"""</a></td> 
    <td>"""+result5[20][0]+"""</td>
    <td>"""+str(100*result5[20][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content59">"""+content59+"""</a></td> 
    <td>"""+result5[21][0]+"""</td>
    <td>"""+str(100*result5[21][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content510">"""+content510+"""</a></td> 
    <td>"""+result5[22][0]+"""</td>
    <td>"""+str(100*result5[22][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content511">"""+content511+"""</a></td> 
    <td>"""+result5[23][0]+"""</td>
    <td>"""+str(100*result5[23][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content512">"""+content512+"""</a></td> 
    <td>"""+result5[24][0]+"""</td>
    <td>"""+str(100*result5[24][1])+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content513">"""+content513+"""</a></td> 
    <td>"""+result5[25][0]+"""</td>
    <td>"""+str(100*result5[25][1])+"""%</td>
    <td></td>
</tr>
"""

    logging.info("Finished Summary 5")

########################################################################################################################
############################################################ Summary 6 #################################################
########################################################################################################################
#Calculate % Passed, ignored 6.3


    perc61=round(100*result62[0][1]/result62[0][2],2)
    perc62=round(100*result62[1][1]/result62[1][2],2)
    perc63=round(100*result63[1]/result63[2],2)
    perc64=round(100*result64[1]/result64[2],2)
    perc65=round(100*result65[1]/result65[2],2)

    calc6=(perc61+perc62+perc63+perc64+perc65)/4
    score6= round(calc6,2)

    summary6 = """
<tr>
    <td><b>6 Networking</b></td> 
    <td></td>
    <td><b>"""+str(score6)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content61">"""+content61+"""</a></td> 
    <td>"""+result62[0][3]+str(result62[0][1])+"""/"""+str(result62[0][2])+"""</td>
    <td>"""+str(perc61)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content62">"""+content62+"""</a></td> 
    <td>"""+result62[1][3]+str(result62[1][1])+"""/"""+str(result62[1][2])+"""</td>
    <td>"""+str(perc62)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content63">"""+content63+"""</a></td> 
    <td>"""+result63[3]+str(result63[1])+"""/"""+str(result63[2])+"""</td>
    <td>"""+str(perc63)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content64">"""+content64+"""</a></td> 
    <td>"""+result64[3]+str(result64[1])+"""/"""+str(result64[2])+"""</td>
    <td>"""+str(perc64)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content65">"""+content65+"""</a></td> 
    <td>"""+result65[3]+str(result65[1])+"""/"""+str(result65[2])+"""</td>
    <td>"""+str(perc65)+"""%</td>
    <td></td>
</tr>
"""
    logging.info("Finished Summary6")

########################################################################################################################
############################################################ Summary 7 #################################################
########################################################################################################################
#Calculate % Passed. 74,75 Not scored

    perc76=round(100*result7[11][0]/result7[11][1],2)
    perc73=round(100*result7[8][0]/result7[8][1],2)
    perc72=round(100*result7[7][0]/result7[7][1],2)
    perc71=round(100*result7[6][0]/result7[6][1],2)

    calc7=(perc71+perc72+perc73+perc76)/4
    score7= round(calc7,2)
    summary7 = """
<tr>
    <td><b>7 Virtual Machines</b></td> 
    <td></td>
    <td><b>"""+str(score7)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content71">"""+content71+"""</a></td> 
    <td>"""+result7[6][2]+str(result7[6][0])+"""/"""+str(result7[6][1])+"""</td>
    <td>"""+str(perc71)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content72">"""+content72+"""</a></td> 
    <td>"""+result7[7][2]+str(result7[7][0])+"""/"""+str(result7[7][1])+"""</td>
    <td>"""+str(perc72)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content73">"""+content73+"""</a></td> 
    <td>"""+result7[8][2]+str(result7[8][0])+"""/"""+str(result7[8][1])+"""</td>
    <td>"""+str(perc73)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content74">"""+content74+"""</a></td> 
    <td>Not Scored</td>
    <td>Not Scored</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content75">"""+content75+"""</a></td> 
    <td>Not Scored</td>
    <td>Not Scored</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content76">"""+content76+"""</a></td> 
    <td>"""+result7[11][2]+str(result7[11][0])+"""/"""+str(result7[11][1])+"""</td>
    <td>"""+str(perc76)+"""%</td>
    <td></td>
</tr>
"""

    logging.info("Finished Summary 7")
########################################################################################################################
############################################################ Summary 8 #################################################
########################################################################################################################
#Calculate % Passed. 83 Not scored

    perc81=round(100*result80[0][1]/result80[0][2],2)
    perc82=round(100*result80[1][1]/result80[1][2],2)

    calc8=(perc81+perc82)/2
    score8= round(calc8,2)
    summary8 = """
<tr>
    <td><b>8 Other Security Considerations</b></td> 
    <td></td>
    <td><b>"""+str(score8)+"""%</b></td>
    <td></td>
</tr>
<tr>
    <td><a href="#content81">"""+content81+"""</a></td> 
    <td>"""+result80[0][3]+str(result80[0][1])+"""/"""+str(result80[0][2])+"""</td>
    <td>"""+str(perc81)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content82">"""+content82+"""</a></td> 
    <td>"""+result80[1][3]+str(result80[1][1])+"""/"""+str(result80[1][2])+"""</td>
    <td>"""+str(perc82)+"""%</td>
    <td></td>
</tr>
<tr>
    <td><a href="#content83">"""+content83+"""</a></td> 
    <td>Not Scored</td>
    <td>Not Scored</td>
    <td></td>
</tr>
"""

    logging.info("Finished Summary 8")
########################################################################################################################
############################################################ Total  ####################################################
########################################################################################################################


    calcfinal=(score2+score3+score4+score5+score6+score7+score8)/7
    finalscore=round(calcfinal,2)
    total = """
<tr>
    <td><b><font color="blue">Total</font></b></td> 
    <td></td>
    <td><b><font color="blue">"""+str(finalscore)+"""%</font></b></td>
</tr>
</table> 
"""

############## detailed Results ##########
    html_content = """
<h1 id="Details">2 - Detailed</h1>
""" + content1+"</ul>"+content2+"</ul>"+content3+"</ul>"+content4+"</ul>"+content5+"</ul>"+content6+"</ul>"+content7+"</ul>"+content8

    html_summary=summary0+summary1+summary2+summary3+summary4+summary5+summary6+summary7+summary8+total
#### Create the HTML File #####
    """
    filename = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    reportname=("CIS-Azure-%s-%s.html" % (name,filename))
    with open(reportname,"w") as cis_result:
        cis_result.write(html_start)
        cis_result.write(html_summary)
        cis_result.write(html_content)
        cis_result.write(html_end)
    logging.info("Report %s was Created" % reportname)
    """

    ################ CSV EXPORT ###################
    sep = ","
    csv_content_hdr = '"SUBSCRIPTION","CHECK","RESULT","DETAILS"\n'

    csv_content1_3  = r(c(content13)+c(result13[3]+str(result13[1])+"""/"""+str(result13[2]))+c(result13[0]))
    csv_content12_3  = r(c(content123)+c(result123[3]+str(result123[1])+"""/"""+str(result123[2]))+c(result123[0]))

    csv_content2_1  = r(c(content22)+c(result22[18][0])+c(result22[0]))   +r(c(content23)+c(result22[19][0])+c(result22[1]))
    csv_content2_2  = r(c(content24)+c(result22[20][0])+c(result22[2]))   +r(c(content25)+c(result22[21][0])+c(result22[3]))     +r(c(content26)+c(result22[22][0])+c(result22[4])) + \
                      r(c(content27)+c(result22[23][0])+c(result22[5]))   +r(c(content28)+c(result22[24][0])+c(result22[6]))
    csv_content2_3  = r(c(content29)+c(result22[25][0])+c(result22[7]))   +r(c(content210)+c(result22[26][0])+c(result22[8]))    +r(c(content211)+c(result22[27][0])+c(result22[9]))
    csv_content2_4  = r(c(content212)+c(result22[28][0])+c(result22[10])) +r(c(content213)+c(result22[29][0])+c(result22[11]))   +r(c(content214)+c(result22[30][0])+c(result22[12]))
    csv_content2_5  = r(c(content215)+c(result22[31][0])+c(result22[13])) +r(c(content216)+c(result22[32][0])+c(result22[14]))   +r(c(content217)+c(result22[33][0])+c(result22[15]))
    csv_content2_6  = r(c(content218)+c(result22[34][0])+c(result22[16])) +r(c(content219)+c(result22[35][0])+c(result22[17]))
    csv_content3_1  = r(c(content31)+c(result31[3]+str(result31[1])+"""/"""+str(result31[2]))+c(result31[0]))
    csv_content3_2  = r(c(content32)+c(result32[3]+str(result32[1])+"""/"""+str(result32[2]))+c(result32[0]))
    csv_content3_3  = r(c(content33)+c(result33[3]+str(result33[1])+"""/"""+str(result33[2]))+c(result33[0]))
    csv_content3_4  = r(c(content34)+c("")+c(result34))                   +r(c(content35)+c("")+c(result35))                     +r(c(content36)+c(result36[3]+str(result36[1])+"""/"""+str(result36[2]))+c(result36[0])) + \
                      r(c(content37)+c(result37[3]+str(result37[1])+"""/"""+str(result37[2]))+c(result37[0]))
    csv_content41_1 = r(c(content411))
    csv_content41_2 = r(c(content414))
    csv_content42_1 = r(c(content421)+c(result42[0][3]+str(result42[0][1])+"""/"""+str(result42[0][2]))+c(result42[0][0]))       +r(c(content422)+c(result42[1][3]+str(result42[1][1])+"""/"""+str(result42[1][2]))+c(result42[1][0])) + \
                      r(c(content423)+c(result42[2][3]+str(result42[2][1])+"""/"""+str(result42[2][2]))+c(result42[2][0]))
    csv_content42_2 = r(c(content424)+c(result42[3][3]+str(result42[3][1])+"""/"""+str(result42[3][2]))+c(result42[3][0]))       +r(c(content425)+c(result42[4][3]+str(result42[4][1])+"""/"""+str(result42[4][2]))+c(result42[4][0])) + \
                      r(c(content426)+c(result42[5][3]+str(result42[5][1])+"""/"""+str(result42[5][2]))+c(result42[5][0]))
    csv_content42_3 = r(c(content427)+c(result42[6][3]+str(result42[6][1])+"""/"""+str(result42[6][2]))+c(result42[6][0]))       +r(c(content428)+c(result42[7][3]+str(result42[7][1])+"""/"""+str(result42[7][2]))+c(result42[7][0]))
    csv_content5_1  = r(c(content51)+c(result5[13][0])+c(result5[0]))     +r(c(content52)+c(result5[14][0])+c(result5[1]))       +r(c(content53)+c(result5[15][0])+c(result5[2]))
    csv_content5_2  = r(c(content54)+c(result5[16][0])+c(result5[3]))     +r(c(content55)+c(result5[17][0])+c(result5[4]))       +r(c(content56)+c(result5[18][0])+c(result5[5]))       +r(c(content57)+c(result5[19][0])+c(result5[6])) + \
                      r(c(content58)+c(result5[20][0])+c(result5[7]))
    csv_content5_3  = r(c(content59)+c(result5[21][0])+c(result5[8]))     +r(c(content510)+c(result5[22][0])+c(result5[9]))      +r(c(content511)+c(result5[23][0])+c(result5[10]))
    csv_content5_4  = r(c(content512)+c(result5[24][0])+c(result5[11]))   +r(c(content513)+c(result5[25][0])+c(result5[12]))
    csv_content6_1  = r(c(content61)+c(result62[0][3]+str(result62[0][1])+"""/"""+str(result62[0][2]))+c(result62[0][0]))        +r(c(content62)+c(result62[1][3]+str(result62[1][1])+"""/"""+str(result62[1][2]))+c(result62[1][0])) + \
                      r(c(content63)+c("")+c(result63[0]))
    csv_content6_2  = r(c(content64)+c(result64[3]+str(result64[1])+"""/"""+str(result64[2]))+c(result64[0]))
    csv_content6_3  = r(c(content65)+c(result65[3]+str(result65[1])+"""/"""+str(result65[2]))+c(result65[0]))
    csv_content7_1  = r(c(content71)+c(result7[6][2]+str(result7[6][0])+"""/"""+str(result7[6][1]))+c(result7[0]))               +r(c(content72)+c(result7[7][2]+str(result7[7][0])+"""/"""+str(result7[7][1]))+c(result7[1])) + \
                      r(c(content73)+c(result7[8][2]+str(result7[8][0])+"""/"""+str(result7[8][1]))+c(result7[2]))
    csv_content7_2  = r(c(content74)+c("")+c(result7[3]))                 +r(c(content75)+c("")+c(result7[4]))                   +r(c(content76)+c(result7[11][2]+str(result7[11][0])+"""/"""+str(result7[11][1]))+c(result7[5]))

    csv_content41   = csv_content41_1 + csv_content41_2
    csv_content42   = csv_content42_1 + csv_content42_2 + csv_content42_3
    csv_content50   = csv_content5_1  + csv_content5_2  + csv_content5_3  + csv_content5_4

    csv_content1 = csv_content1_3 + csv_content12_3
    csv_content2 = csv_content2_1 + csv_content2_2 + csv_content2_3 + csv_content2_4 + csv_content2_5 + csv_content2_6
    csv_content3 = csv_content3_1 + csv_content3_2 + csv_content3_3 + csv_content3_4
    csv_content4 = csv_content41  + csv_content42
    csv_content5 = csv_content50
    csv_content6 = csv_content6_1 + csv_content6_2 + csv_content6_3
    csv_content7 = csv_content7_1 + csv_content7_2
    csv_content8 = r(c(content81)+c(result80[0][3]+str(result80[0][1])+"""/"""+str(result80[0][2]))+c(result80[0][0])) + r(c(content82)+c(result80[1][3]+str(result80[1][1])+"""/"""+str(result80[1][2]))+c(result80[1][0])) + \
                   r(c(content83)+c("")+c(result83))

    reportname = tempfile.gettempdir() + os.path.sep + 'azure_cis_bench_out.csv'
    with open(reportname,"a") as cis_result:
        cis_result.write(csv_content_hdr)
        cis_result.write(strip_tags(csv_content1))
        cis_result.write(strip_tags(csv_content2))
        cis_result.write(strip_tags(csv_content3))
        cis_result.write(strip_tags(csv_content4))
        cis_result.write(strip_tags(csv_content5))
        cis_result.write(strip_tags(csv_content6))
        cis_result.write(strip_tags(csv_content7))
        cis_result.write(strip_tags(csv_content8))

def run_tests():
    reportname = tempfile.gettempdir() + os.path.sep + 'azure_cis_bench_out.csv'
    if os.path.isfile(reportname):
        os.remove(reportname)

    try:
        subs=az_cis_utils.get_subscriptions()

        if subs is None:
            logging.error("No Azure subscriptions found...exiting")
            sys.exit(1)

        logging.info("Running Azure CIS checks")

        for j in range(0,len(subs[0])):
            logging.info("Starting CIS Benchmark for: %s", subs[2][j])
            os.popen('az account set --subscription "%s"' % subs[2][j])
            generate_report(subs[0][j],subs[2][j],subs[1][j])

        logging.info("Done")
        return reportname
    except Exception as e:
        logging.error("Exception in main: %s %s", type(e), str(e.args))
        sys.exit(1)




import sys
import json
import os
import requests
import re
import logging

gAllVMs = None

def get_all_vms(params):
    global gAllVMs
    if gAllVMs is not None:
        return gAllVMs
    gAllVMs = { }
    access_token = params['access_token']
    allsubs = get_all_subscriptions(access_token)
    for sub in allsubs:
        resourcegroups = get_all_resourcegroups_for_subscription(sub, access_token)
        for res_group in resourcegroups:
            vms = get_vms(access_token, sub, res_group)
            gAllVMs.update(vms)
    return gAllVMs

# Prints details about subscriptions, resource groups and workspaces
def print_details(token):
    allsubs = get_all_subscriptions(token)
    print( "")
    print("Missing details for subscription/resource group/workspace....")
    print("Available subscriptions with resource group and workspace details as below:")
    for sub in allsubs:
        print("Subscription: %s" % sub)
        resourcegroups = get_all_resourcegroups_for_subscription(sub, token)
        for res_group in resourcegroups:
            print(" ** Resource group: %s" % res_group)
        workspaces = get_all_workspaces_for_subscription(sub, token)
        for workspace in workspaces:
            print(" ** Workspace: %s" % workspace)
    print("")
    print("Please re-run twigs with appropriate values for subscription, resource group and workspace.")
    print("")
    sys.exit(1)

# Main entry point
def get_inventory(args):
    params =  {}
    params['handle'] = args.handle
    params['tenant_id'] = args.azure_tenant_id
    params['app_id'] = args.azure_application_id
    params['app_key'] = args.azure_application_key
    params['subscription'] = args.azure_subscription
    params['resource_group'] = args.azure_resource_group
    params['workspace'] = args.azure_workspace
    params['enable_tracking_tags'] = args.enable_tracking_tags
    token = get_access_token(params)
    if token is None:
        return
    params['access_token'] = token
    get_all_vms(params)
    if args.azure_subscription is None or args.azure_resource_group is None or args.azure_workspace is None:
        print_details(token)
        return
    assets = retrieve_inventory(params)
    return assets

def parse_inventory(email,data,params):
    logging.info("Processing inventory retrieved from Azure...")
    hosts = []
    assets = []
    asset_map = {}
    for i in range(len(data)):
        if data[i][4] == 'WindowsServices': #ConfigDataType
            continue
        host = data[i][5]
        vmuuid = data[i][6]
        publisher = data[i][2]

        if host not in hosts  and publisher != '0':
            patches = []
            products = []
            asset_map = {}
            asset_map['owner'] = email
            asset_map['host'] = host
            asset_map['id'] = host
            asset_map['name'] = host
            asset_map['tags'] = [ ]
            asset_map['patch_tracker'] = { } # To help remove duplicate patches
            asset_map['vmuuid'] = vmuuid
            if data[i][1] == 'Update': #ApplicationType for MS patches
                patch = parse_patch(data[i])
                if patch is not None:
                    patches.append(patch)
                    asset_map['patch_tracker'][patch['id']] = patch['id']
            if data[i][1] == 'Package' or data[i][1] == 'Application': #ApplicationType for Linux packages
                pname = data[i][0]
                pversion =  data[i][3]
                products.append(pname+' ' + pversion)
            asset_map['products'] = products
            asset_map['patches'] = patches
            os, os_version = get_os_details(host, vmuuid, params)
            asset_map['type'] = get_os_type(os)
            if len(asset_map['type']) > 0:
                asset_map['tags'].append(asset_map['type'])
            if asset_map['type'] == 'Windows':
                asset_map['tags'].append('OS_RELEASE:' + os)
                asset_map['tags'].append('OS_VERSION:' + os_version)
            else:
                asset_map['tags'].append('OS_RELEASE:%s %s' % (os, os_version))
            if params['enable_tracking_tags'] == True:
                asset_map['tags'].append("SOURCE:Azure:" + params['tenant_id'])
            else:
                asset_map['tags'].append("SOURCE:Azure")
            assets.append(asset_map)
            hosts.append(host)
        else:
            for asset in assets:
                if asset['host'] == host:
                    products = asset['products']
                    patches = asset['patches']
                    if data[i][1] == 'Update': #ApplicationType for MS patches
                        patch = parse_patch(data[i])
                        if patch is not None and asset['patch_tracker'].get(patch['id']) is None:
                            patches.append(patch)
                            asset['patches'] = patches
                            asset['patch_tracker'][patch['id']] = patch['id']
                    if data[i][1] == 'Package' or data[i][1] == 'Application': #ApplicationType for Linux packages
                        pname = data[i][0]
                        pversion =  data[i][3]
                        products.append(pname+' ' + pversion)
                        asset['products'] = products
    # Remove the additional fields 'patch_tracker' (added to avoid duplicate patches) & 'vmuuid'
    for asset in assets:
        asset.pop('patch_tracker', None)
        asset.pop('vmuuid', None)
    return assets

def parse_patch(data):
    patch_id = re.findall(r'(KB[0-9]+)', data[0])
    if len(patch_id) == 0:
        return None
    patch = {}
    patch['url'] = ''
    patch['id'] = patch_id[0]
    patch['product'] = ''
    patch['description'] = data[0]
    return patch
            
def get_os_type(ostype):
    if ostype is None:
        return ''
    elif 'Microsoft' in  ostype or 'Windows' in ostype:
        return 'Windows'
    elif 'Red Hat' in ostype or 'redhat' in ostype:
        return 'Red Hat'
    elif 'Ubuntu' in ostype or 'ubuntu' in ostype:
        return 'Ubuntu'
    elif 'Oracle' in ostype or 'oracle' in ostype:
        return "Oracle Linux"
    return ''

def retrieve_inventory(params):
    email = params['handle']
    sub_id = params['subscription']
    resource_group = params['resource_group']
    workspace_id = params['workspace']
    token = params['access_token']
    headers = { "Content-Type":"application/json", "Authorization": "Bearer %s" % token }
    url = 'https://management.azure.com/subscriptions/%s/resourceGroups/%s/providers/Microsoft.OperationalInsights/workspaces/%s/api/query?api-version=2017-01-01-preview' % (sub_id,resource_group,workspace_id)
    json_data = {"query":"ConfigurationData | summarize by SoftwareName, SoftwareType, Publisher, CurrentVersion, ConfigDataType, Computer, VMUUID"}

    logging.info("Retrieving inventory details from Azure...") 

    resp = requests.post(url, headers=headers, json=json_data)
    if resp.status_code == 200:
        response = resp.json()
        if response.get('Tables'):
            tables = response['Tables']
            return parse_inventory(email,tables[0]['Rows'],params)
    else:
        logging.error("Error could not get asset inventory details from Azure...")
        logging.error("Response content: %s" % resp.text)
        sys.exit(1)

#Get access token using  an AAD, an app id associted with that AAD and the API key/secret for that app
def get_access_token(params):         
    aad_id = params['tenant_id']
    aad_app_id = params['app_id']
    app_key = params['app_key']
    url = "https://login.microsoftonline.com/" + aad_id + "/oauth2/token"

    logging.info("Getting access token...") 

    resp = requests.post(url, data={"grant_type":"client_credentials", "client_id": aad_app_id, "client_secret": app_key, "resource":"https://management.azure.com/"})
    if resp.status_code == 200:
        response = resp.json()
        token = response['access_token']
    else:
        logging.error("Error unable to get access token for API calls")
        logging.error("Response content: %s" % resp.text)
        sys.exit(1)

    return token

# Try to get OS details for given VM
def get_os_details(host, vmuuid, params):
    all_vms = get_all_vms(params)
    vm_id = all_vms.get(vmuuid)
    if vm_id is None:
        # Handle Endian-ness issue with Azure VM UUIDs
        tokens = vmuuid.split('-')
        t = tokens[0]
        alt_vmuuid = "%s%s%s%s" % (t[6:8], t[4:6], t[2:4], t[0:2])
        t = tokens[1]
        alt_vmuuid = alt_vmuuid + "-%s%s" % (t[2:4], t[0:2])
        t = tokens[2]
        alt_vmuuid = alt_vmuuid + "-%s%s" % (t[2:4], t[0:2])
        alt_vmuuid = alt_vmuuid + '-%s-%s' % (tokens[3], tokens[4])
        vm_id = all_vms.get(alt_vmuuid)
        if vm_id is None:
            return None, None

    headers = { "Content-Type":"application/json", "Authorization": "Bearer %s" % params['access_token'] }
    url = "https://management.azure.com" + vm_id + "/instanceView?api-version=2018-06-01"

    resp = requests.get(url, headers=headers)
    if resp.status_code == 200:
        response = resp.json()
        return response.get('osName'), response.get('osVersion')
    else:
        logging.warn("Warning unable to get OS version details for VM [%s]. It might not be running..." % host)
        return None, None

# Get details for all VMs
def get_vms(access_token, subscription, resource_group):
    all_vms = { }
    headers = { "Content-Type":"application/json", "Authorization": "Bearer %s" % access_token }
    url = "https://management.azure.com/subscriptions/%s/resourceGroups/%s/providers/Microsoft.Compute/virtualMachines?api-version=2019-12-01" % (subscription, resource_group)
    while url is not None:
        resp = requests.get(url, headers=headers)
        if resp.status_code == 200:
            response = resp.json()
            for vm in response['value']:
                vmuuid = vm['properties']['vmId']
                all_vms[vmuuid] = vm['id']
            url = response.get('nextLink')
        else:
           url = None
    return all_vms

# Get a list of subscriptions for current AAD/Tenant
def get_all_subscriptions(token):
    allsubs = []
    headers = { "Content-Type":"application/json", "Authorization": "Bearer %s" % token }
    url = "https://management.azure.com/subscriptions?api-version=2018-01-01"

    resp = requests.get(url, headers=headers)

    if resp.status_code == 200:
        subs = resp.json()
        sublist = subs['value']
        for sub in sublist:
            allsubs.append(sub['subscriptionId'])
    else:
        logging.error("API call to get all subscriptions failed")
        logging.error("Response content: %s" % resp.text)
        sys.exit(1)
    return allsubs

#Get all resource groups for a subscription
def get_all_resourcegroups_for_subscription(subid,token):
    allresourcegroups = []
    headers = { "Content-Type":"application/json", "Authorization": "Bearer %s" % token }
    url = 'https://management.azure.com/subscriptions/%s/resourcegroups?api-version=2018-01-01' % subid
    resp = requests.get(url, headers=headers)
    if resp.status_code == 200:
        rgroups = resp.json()
        rlist = rgroups['value']
        for r in rlist:
            allresourcegroups.append(r['name'])
    else:
        logging.error("API call to get all resource groups failed")
        logging.error("Response content: %s" % resp.text)
        sys.exit(1)

    return allresourcegroups

#Get all workspaces for the subscription
def get_all_workspaces_for_subscription(subid,token):
    allworkspaces = []
    headers = { "Content-Type":"application/json", "Authorization": "Bearer %s" % token }
    url = 'https://management.azure.com/subscriptions/%s/providers/Microsoft.OperationalInsights/workspaces?api-version=2017-01-01-preview' % subid

    resp = requests.get(url, headers=headers)
    if resp.status_code == 200:
        workspaces = resp.json()
        wlist = workspaces['value']
        for w in wlist:
            allworkspaces.append(w['name'])
    else:
        logging.error("API call to get all workspaces failed")
        logging.error("Response content: %s" % resp.text)
        sys.exit(1)

    return allworkspaces




import sys
import platform
import os
import subprocess
import logging
from xml.dom.minidom import parse, parseString
import csv
from . import linux

NMAP = "/usr/bin/nmap"

def nmap_exists():
    return os.path.isfile(NMAP) and os.access(NMAP, os.X_OK)

def discover(args):
    handle = args.handle
    token = args.token
    instance = args.instance

    hosts = args.hosts

def get_inventory(args):
    if not nmap_exists():
        logging.error('nmap CLI not found')
        return None

    logging.info("Fingerprinting "+args.hosts)
    cmdarr = [NMAP + ' -oX - -sV -PN -T4 -F '+args.hosts]
    try:
        out = subprocess.check_output(cmdarr, shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error determining OS release")
        return None 

    assets = []
    dom = parseString(out)
    hosts = dom.getElementsByTagName("host")
    for h in hosts:
        addr = h.getElementsByTagName("address")[0]
        addr = addr.getAttribute('addr')
        hostname = addr
        harr = h.getElementsByTagName("hostname")
        if harr != None and len(harr) > 0:
            hostname = h.getElementsByTagName("hostname")[0]
            hostname = hostname.getAttribute('name')
        cpes = h.getElementsByTagName("cpe")
        products = []
        for c in cpes:
            cstr = c.firstChild.data
            carr = cstr.split(':')
            prodstr = carr[2] + ' ' + carr[3] + ' '
            if len(carr) >= 5:
                prodstr += carr[4]
            prodstr = prodstr.strip()
            prodstr = prodstr.replace('_',' ')
            if prodstr not in products:
                products.append(prodstr)
        asset_data = {}
        asset_data['id'] = addr 
        asset_data['name'] = hostname 
        asset_data['type'] = 'Other' 
        asset_data['owner'] = args.handle
        asset_data['products'] = products 
        asset_tags = []
        asset_data['tags'] = asset_tags
        if args.no_ssh_audit == False:
            ssh_issues = linux.run_ssh_audit(args, addr, addr)
            if len(ssh_issues) != 0:
                asset_data['tags'].append('SSH Audit')
            asset_data['config_issues'] = ssh_issues
        assets.append(asset_data)
    return assets



import sys
import logging
from . import gcp_cis_utils as gcp_cis_utils

def check4_1():
    # 4.1 Ensure that instances are not configured to use the default service account (Scored)

    logging.info("4.1 Ensure that instances are not configured to use the default service account (Scored)")
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        out_json = gcp_cis_utils.run_gcloud_cmd("projects describe %s" % p)
        project_no = out_json['projectNumber']
        default_service_account = project_no + "-compute@developer.gserviceaccount.com"
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            if (entry['name'].startswith('gke-') and entry.get('labels') is not None and 'goog-gke-node' in entry['labels'].keys()) or (entry.get('labels') is not None and 'cloud.google.com/gke-nodepool' in entry['labels'].keys()):
                continue
            if entry.get('serviceAccounts') is None:
                continue
            for sa in entry['serviceAccounts']:
                if sa['email'] == default_service_account:
                    details.append("Compute instance [%s] uses default service account [%s] in project [%s]" % (entry['name'], default_service_account, p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.1', '4.1 [Level 1] Ensure that instances are not configured to use the default service account (Scored)', "\n".join(details), '4', '', '')
    return None

def check4_2():
    # 4.2 Ensure that instances are not configured to use the default service account with full access to all Cloud APIs (Scored)

    logging.info("4.2 Ensure that instances are not configured to use the default service account with full access to all Cloud APIs (Scored)")
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        out_json = gcp_cis_utils.run_gcloud_cmd("projects describe %s" % p)
        project_no = out_json['projectNumber']
        default_service_account = project_no + "-compute@developer.gserviceaccount.com"
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            if (entry['name'].startswith('gke-') and entry.get('labels') is not None and 'goog-gke-node' in entry['labels'].keys()) or (entry.get('labels') is not None and 'cloud.google.com/gke-nodepool' in entry['labels'].keys()):
                continue
            if entry.get('serviceAccounts') is None:
                continue
            for sa in entry['serviceAccounts']:
                if sa['email'] == default_service_account and "https://www.googleapis.com/auth/cloud-platform" in sa['scopes']:
                    details.append("Compute instance [%s] uses default service account [%s] with full access to all Cloud APIs in project [%s]" % (entry['name'], default_service_account, p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.2', '4.2 [Level 1] Ensure that instances are not configured to use the default service account with full access to all Cloud APIs (Scored)', "\n".join(details), '4', '', '')
    return None

def check4_3():
    # 4.3 Ensure "Block Project-wide SSH keys" is enabled for VM instances (Scored)

    logging.info('4.3 Ensure "Block Project-wide SSH keys" is enabled for VM instances (Scored)')
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            if (entry['name'].startswith('gke-') and entry.get('labels') is not None and 'goog-gke-node' in entry['labels'].keys()) or (entry.get('labels') is not None and 'cloud.google.com/gke-nodepool' in entry['labels'].keys()):
                continue
            metadata = entry.get('metadata')
            block_pw_ssh_keys_enabled = False
            if metadata is not None:
                metadata_items = metadata.get('items')
                if metadata_items is not None:
                    for mdi in metadata_items:
                        if mdi['key'] == 'block-project-ssh-keys' and mdi['value'] == 'true':
                            block_pw_ssh_keys_enabled = True
            if block_pw_ssh_keys_enabled == False:
                details.append("Compute instance [%s] does not have Block Project-wide SSH keys enabled in project [%s]" % (entry['name'], p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.3', '4.3 [Level 1] Ensure "Block Project-wide SSH keys" is enabled for VM instances (Scored)', "\n".join(details), '4', '', '')
    return None

def check4_4():
    # 4.4 Ensure oslogin is enabled for a Project (Scored)

    logging.info('4.4 Ensure oslogin is enabled for a Project (Scored)')
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        oslogin_enabled = False
        out_json = gcp_cis_utils.run_gcloud_cmd("compute project-info describe --project=%s" % p)
        cimd = out_json.get('commonInstanceMetadata')
        if cimd is not None:
            metadata_items = cimd.get('items')
            if metadata_items is not None:
                for mdi in metadata_items:
                    if mdi['key'] == 'enable-oslogin' and mdi['value'] == 'TRUE':
                        oslogin_enabled = True
        overriding_instances = []
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            metadata = entry.get('metadata')
            instance_overrides = False
            if metadata is not None:
                metadata_items = metadata.get('items')
                if metadata_items is not None:
                    for mdi in metadata_items:
                        if mdi['key'] == 'enable-oslogin' and mdi['value'] == 'FALSE':
                            instance_overrides = True
            if instance_overrides:
                overriding_instances.append(entry['name'])
        if oslogin_enabled == False:
            details.append("oslogin is not enabled for project [%s]" % p)
        elif len(overriding_instances) > 0:
            details.append("oslogin is enabled for project [%s], however following compute instances override the project setting [%s]" % (p, ",".join(overriding_instances)))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.4', '4.4 [Level 1] Ensure oslogin is enabled for a Project (Scored)', "\n".join(details), '4', '', '')
    return None

def check4_5():
    # 4.5 Ensure 'Enable connecting to serial ports' is not enabled for VM Instance (Scored)

    logging.info("4.5 Ensure 'Enable connecting to serial ports' is not enabled for VM Instance (Scored)")
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            metadata = entry.get('metadata')
            serial_port_enabled = False
            if metadata is not None:
                metadata_items = metadata.get('items')
                if metadata_items is not None:
                    for mdi in metadata_items:
                        if mdi['key'] == 'serial-port-enable' and mdi['value'] == 'true':
                            serial_port_enabled = True
            if serial_port_enabled:
                details.append("Serial port is enabled for compute instance [%s] in project [%s]" % (entry['name'], p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.5', "4.5 [Level 1] Ensure 'Enable connecting to serial ports' is not enabled for VM Instance (Scored)", "\n".join(details), '4', '', '')
    return None

def check4_6():
    # 4.6 Ensure that IP forwarding is not enabled on Instances (Scored)

    logging.info("4.6 Ensure that IP forwarding is not enabled on Instances (Scored)")
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            if (entry['name'].startswith('gke-') and entry.get('labels') is not None and 'goog-gke-node' in entry['labels'].keys()) or (entry.get('labels') is not None and 'cloud.google.com/gke-nodepool' in entry['labels'].keys()):
                continue
            if entry.get('canIpForward') is not None and entry['canIpForward']:
                details.append("IP forrwarding is enabled on compute instance [%s] in project [%s]" % (entry['name'], p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.6', '4.6 [Level 1] Ensure that IP forwarding is not enabled on Instances (Scored)', "\n".join(details), '4', '', '')
    return None

def check4_7():
    # 4.7 Ensure VM disks for critical VMs are encrypted with Customer- Supplied Encryption Keys (CSEK) (Scored)

    logging.info("Not supported - 4.7 Ensure VM disks for critical VMs are encrypted with Customer- Supplied Encryption Keys (CSEK) (Scored)")
    return None

def check4_8():
    # 4.8 Ensure Compute instances are launched with Shielded VM enabled (Scored)

    logging.info("4.8 Ensure Compute instances are launched with Shielded VM enabled (Scored)")
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            shieldedInstanceConfig = entry.get('shieldedInstanceConfig')
            if shieldedInstanceConfig is None:
                details.append("Compute instance [%s] in project [%s] is not launched with Shielded VM enabled" % (entry['name'], p))
            else:
                if shieldedInstanceConfig['enableVtpm'] == False or shieldedInstanceConfig['enableIntegrityMonitoring'] == False:
                    details.append("Compute instance [%s] in project [%s] is not launched with Shielded VM enabled" % (entry['name'], p))
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.8', '4.8 [Level 2] Ensure Compute instances are launched with Shielded VM enabled (Scored)', "\n".join(details), '5', '', '')
    return None

def check4_9():
    # 4.9 Ensure that Compute instances do not have public IP addresses (Scored)

    logging.info("4.9 Ensure that Compute instances do not have public IP addresses (Scored)")
    details = []
    compute_instances_by_projects = gcp_cis_utils.get_compute_instances_by_projects()
    for p in compute_instances_by_projects.keys():
        out_json = compute_instances_by_projects[p]
        for entry in out_json:
            if (entry['name'].startswith('gke-') and entry.get('labels') is not None and 'goog-gke-node' in entry['labels'].keys()) or (entry.get('labels') is not None and 'cloud.google.com/gke-nodepool' in entry['labels'].keys()):
                continue
            for ni in entry['networkInterfaces']:
                if ni.get('accessConfigs') is not None:
                    details.append("Compute instance [%s] in project [%s] has Public IP address" % (entry['name'], p))
                    break
    if len(details) > 0:
        return gcp_cis_utils.create_issue('cis-gcp-bench-check-4.9', '4.9 [Level 2] Ensure that Compute instances do not have public IP addresses (Scored)', "\n".join(details), '5', '', '')
    return None

def check4_10():
    # 4.10 Ensure that App Engine applications enforce HTTPS connections (Not Scored)

    logging.info("Not supported - 4.10 Ensure that App Engine applications enforce HTTPS connections (Not Scored)")
    return None

def run_checks():
    config_issues = []
    gcp_cis_utils.append_issue(config_issues, check4_1())
    gcp_cis_utils.append_issue(config_issues, check4_2())
    gcp_cis_utils.append_issue(config_issues, check4_3())
    gcp_cis_utils.append_issue(config_issues, check4_4())
    gcp_cis_utils.append_issue(config_issues, check4_5())
    gcp_cis_utils.append_issue(config_issues, check4_6())
    gcp_cis_utils.append_issue(config_issues, check4_7())
    gcp_cis_utils.append_issue(config_issues, check4_8())
    gcp_cis_utils.append_issue(config_issues, check4_9())
    gcp_cis_utils.append_issue(config_issues, check4_10())
    return config_issues

